# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.


# Write file to perform conformational search
def make_confsearch_script(scriptname, pickle_file, smiles, path='', iterations=2000,
							RMSthresh=10.0, maxconfs=10, Ethresh=999.99):
	# Input:
	# scriptname : name of script to produce
	# pickle_file: pickle file containing molecule (relative to 'path')
	# smiles: smiles string representing molecule
	# aE_dir: directory containing
	# path: path to molecule folder
	# iterations: number of iterations for torsional angle search
	# RMSthresh: RMS threshold for torsional angle search
	# maxconfs: Maximum number of conformers to produce
	# Ethresh: Energy threshold above which conformers are discarded

	# Returns: None

	strings = []
	# produce script to do conformational search
	strings.append('import sys')
	strings.append("import autoENRICH')")
	strings.append('import pickle')
	strings.append('from file_creation.structure_formats import xyz')
	strings.append("file = '{0:<10s}'".format(pickle_file.split('/')[-1]))
	strings.append("molecule = pickle.load(open(file, 'rb'))")
	strings.append("molecule.generate_conformers('{smiles:<10s}', path='{path:<10s}', iterations={its:<10d}".format(smiles=smiles,
																												path=path,
																												its=iterations))
	strings.append(", RMSthresh={RMS:<10f}, maxconfs={maxconfs:<10d}, Ethresh={Ethresh:<10f})".format(RMS=RMSthresh,
																									maxconfs=maxconfs,
																									Ethresh=Ethresh))
	strings.append("for conformer in molecule.conformers:")
	strings.append("\txyz_file = 'conf_search/' +  conformer.molid + '.xyz'")
	strings.append("\txyz.nmrmol_to_xyz(molecule, xyz_file)")
	strings.append("pickle.dump(molecule, open(file, 'wb'))")

	# print script to file
	with open(path + scriptname, 'w') as f:
		for string in strings:
			print(string, file=f)



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

def Get_optimal_HP(rep, parameter='HCS', dataset='set3'):
	optimal = {
					'set2_all' :	{'cmat' : {	'HCS' : {'sigma':100, 'lambda':1e-14, 'cutoff':2.0},
												'CCS' : {'sigma':1e+10, 'lambda':1e-8, 'cutoff':3.0},
												'J1CH' : {'sigma':54555.9, 'lambda':0.0001, 'cutoff':3.0}},

									'slatm'	: { 'HCS' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0},
												'CCS' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0},
												'J1CH' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0}},

									'fchl'	: {	'HCS' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0},
												'CCS' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0}}},

					'set2_hcno' :	{'cmat' : {	'HCS' : {'sigma':1e+10, 'lambda':1e-8, 'cutoff':3.0},
												'CCS' : {'sigma':1e+10, 'lambda':1e-8, 'cutoff':3.0},
												'J1CH' : {'sigma':2e+09, 'lambda':1e-08, 'cutoff':3.0}},

									'slatm'	: { 'HCS' : {'sigma':177828, 'lambda':1e-8, 'cutoff':7.0},
												'CCS' : {'sigma':177828, 'lambda':1e-8, 'cutoff':7.0},
												'J1CH' : {'sigma':1624, 'lambda':1e-12, 'cutoff':4.0}},

									'fchl'	: {	'HCS' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0},
												'CCS' : {'sigma':10, 'lambda':1e-8, 'cutoff':5.0}}},

					'set3' :		{'cmat' : {	'HCS' : {'sigma':2.511e+7, 'lambda':1e-6, 'cutoff':3.0}, #0.53
												'CCS' : {'sigma':1e+10, 'lambda':1e-8, 'cutoff':3.0},
												'J1CH' : {'sigma':119356, 'lambda':1.1e-05, 'cutoff':1.82}}, #2.63

									'slatm'	: { 'HCS' : {'sigma':30683.8, 'lambda':3.9e-05, 'cutoff':7.52}, # 0.45
												'CCS' : {'sigma':13.4951, 'lambda':1.8e-16, 'cutoff':6.88}, # 4.04
												'J1CH' : {'sigma':8.503e+06, 'lambda':2.5e-08, 'cutoff':6.49}}, #1.53

									'fchl'	: {	'HCS' : {'sigma':0.431, 'lambda':0.0068, 'cutoff':4.67}, # 0.37
												'CCS' : {'sigma':0.352, 'lambda':0.0023, 'cutoff':7.82}, # 3.87
												'J1CH' : {'sigma':1.414, 'lambda':9.3e-06, 'cutoff':9.93}}},

					'set4' :		{'cmat' : {	'HCS' : {'sigma':2.511e+7, 'lambda':1e-6, 'cutoff':3.0}, #0.53
												'CCS' : {'sigma':1e+10, 'lambda':1e-8, 'cutoff':3.0},
												'J1CH' : {'sigma':1.66, 'lambda':1.5e-10, 'cutoff':1.66}}, #1.63

									'slatm'	: { 'HCS' : {'sigma':30683.8, 'lambda':3.9e-05, 'cutoff':7.52}, # 0.45
												'CCS' : {'sigma':13.4951, 'lambda':1.8e-16, 'cutoff':6.88}, # 4.04
												'J1CH' : {'sigma':26.6671, 'lambda':7.3e-12, 'cutoff':9.28}}, #0.94

									'fchl'	: {	'HCS' : {'sigma':5.92947, 'lambda':1.4e-06, 'cutoff':6.26}, # 0.26
												'CCS' : {'sigma':0.210613, 'lambda':0.0045, 'cutoff':5.06}, # 3.04
												'1JCH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12}, # 0.94
												'1JNH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12},
												'2JCH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12},
												'2JNH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12},
												'2JHH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12},
												'3JCH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12},
												'3JNH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12},
												'3JHH' : {'sigma':2.23, 'lambda':7.7e-07, 'cutoff':5.12}}}}




	sigma = optimal[dataset][rep][parameter]['sigma']
	lamda = optimal[dataset][rep][parameter]['lambda']
	cutoff = optimal[dataset][rep][parameter]['cutoff']

	return sigma, lamda, cutoff



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import pickle
from autoENRICH.file_creation.HPC_submission import make_HPC_header
from autoENRICH.util.filename_utils import get_unique_part

import sys
import numpy as np
import glob

def setup_predict(args):

	# NEED TO MAKE GPU VERSION
	python_env = args['python_env']
	system = args['system']
	nodes = 1
	ppn = args['processors']
	walltime = args['walltime']
	mem = args['memory']

	jobname = 'IMP_' + 'predict'
	submission_file = jobname + '.submit'
	header = make_HPC_header(jobname, system=args['system'], nodes=1, ppn=args['processors'], walltime=args['walltime'], mem=args['memory'])
	strings = []
	strings.append('')
	strings.append('shopt -s expand_aliases')
	strings.append('source ~/.bashrc')
	strings.append("source activate {env:<10s}".format(env=python_env))
	strings.append("IMPRESSION predict --prefs {0:<10s}".format(args['prefs']))


	with open(submission_file, 'w') as f:
		for string in header:
			print(string, file=f)
		for string in strings:
			print(string, file=f)


def predict(args):

	from autoENRICH.molecule.dataset import dataset
	from autoENRICH.file_creation.structure_formats.nmredata import nmrmol_to_nmredata

	for files_set in args['test_sets']:
		parts = files_set.split('/')
		path = ''
		for part in parts[:-1]:
			path = path + part + '/'

		files = glob.glob(files_set)
		#if len(files) == 0:
		#	print ('No file(s) found matching ', args['training_set'])
		#	sys.exit(0)
		dset = dataset()

		label_part = get_unique_part(files)
		dset.get_mols(files, type='nmredata', label_part=label_part)
		if len(dset.mols) == 0:
			print('No molecules loaded. . .')
			sys.exit(0)

		for m, model_file in enumerate(args['models']):

			print('Predicting from model: ', model_file)

			model = pickle.load(open(model_file, 'rb'))

			print(model.args["targetflag"])

			dset.get_features_frommols(model.args, params=model.params)
			assert len(dset.x) > 0, print('No features made. . . ')

			if args['store_datasets']:
				pickle.dump(dset, open('OPT_testing_set.pkl', 'wb'))

			y_pred = model.predict(dset.x)
			assert len(y_pred) == len(dset.y)

			v_preds = []
			for i in range(args['var']):
				var_model_file = model_file.split('.pkl')[0] + '_' + str(i+1) + '.pkl'

				try:
					var_model = pickle.load(open(var_model_file, 'rb'))
				except Exception as e:
					print(e)
					continue

				assert model.args['featureflag'] == var_model.args['featureflag']
				assert model.args['targetflag'] == var_model.args['targetflag']
				assert model.args['max_size'] == var_model.args['max_size']
				assert model.params == var_model.params, print(model.params, var_model.params)

				print('\tPredicting from ', var_model_file)
				tmp_preds = var_model.predict(dset.x)
				v_preds.append(tmp_preds)

			if args['var'] > 0:
				var = np.var(np.asarray(v_preds), axis=0)
			else:
				var = np.zeros(len(y_pred), dtype=np.float64)

			if m == 0:
				dset.assign_from_ml(y_pred, var, zero=True)
			else:
				dset.assign_from_ml(y_pred, var, zero=False)

		for mol in dset.mols:
			outname = args['output_dir'] + 'IMP_' + mol.molid + '.nmredata.sdf'
			nmrmol_to_nmredata(mol, outname)

	print('Done')



















#####



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from autoENRICH.ml.models.model import genericmodel
import qml
from sklearn.model_selection import KFold
import copy
import sys

class KRRmodel(genericmodel):

	def __init__(self, id='KRRmodel', x=[], y=[], params={}, model_args={}):
		genericmodel.__init__(self, id, x, y, params, model_args)
		self.alpha = []


	def train(self, train_x=[], train_y=[]):
		k = []

		if len(train_x) == 0:
			train_x = self.train_x
		if len(train_y) == 0:
			train_y = self.train_y

		try:
			dimensions = train_x.shape[1]
		except:
			print('training reps not in numpy array (KRRmodel.py, train, l26)')
			train_x = np.asarray(train_x)
			train_y = np.asarray(train_y)

		# reshape x array:
		Xr = []
		for _ in range(dimensions):
			Xr.append([])
		for x in train_x:
			for i in range(dimensions):
				Xr[i].append(x[i])
		Xr = np.asarray(Xr)

		# loop through representation sets in training set
		for d in range(dimensions):
			# append kernel for each representation set
			k.append(qml.kernels.laplacian_kernel(Xr[d], Xr[d], self.params['sigma']) + self.params['lamda'] * np.identity(Xr[d].shape[0]))

		# loop through kernels
		K = k[0]
		if len(k) > 1:
			for i in range(1, len(k)):
				# multiply kernels so result is k1 * k2 * k3 ...
				K = K * k[i]

		# get the KRR prefactors, i.e. this is the training of the network
		self.alpha = qml.math.cho_solve(K, train_y)

		# report training state
		self.trained = True


	def predict(self, test_x, train_x=[]):

		if len(train_x) == 0:
			train_x = self.train_x

		ks = []
		assert test_x.shape[1] == train_x.shape[1]

		dimensions = train_x.shape[1]

		# reshape x arrays:
		Xe = []
		Xr = []
		for _ in range(dimensions):
			Xe.append([])
			Xr.append([])
		for x in test_x:
			for i in range(dimensions):
				Xe[i].append(x[i])
		for x in train_x:
			for i in range(dimensions):
				Xr[i].append(x[i])
		Xe = np.asarray(Xe)
		Xr = np.asarray(Xr)


		# loop through representation sets in training set
		for d in range(dimensions):
			# append kernel for each representation set
			ks.append(qml.kernels.laplacian_kernel(Xe[d], Xr[d], self.params['sigma']))

		# loop through kernels
		Ks = ks[0]
		if len(ks) > 1:
			for i in range(1, len(ks)):
				# multiply kernels so result is k1 * k2 * k3 ...
				Ks = Ks * ks[i]

		# predict values of y
		y_pred = np.dot(Ks, self.alpha)

		return y_pred



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

# define functions to produce QML based features
import qml.fchl
import qml.representations
from autoENRICH.util.flag_handler.hdl_targetflag import flag_to_target
import numpy as np


# all features for QML are of the shape: [n, p, r]
# where n is the number of "features", each n is a single NMR environment
# p is the number of sub environment features needed in each NMR environment
# r is the representation itself

# Make FCHL features
def get_FCHL_features(mols, targetflag='CCS', cutoff=5.0, max=400):
	# Input:
	#	mols: list of autoENRICH nmrmol objects
	#	targetflag: flag corresponding to nmr parameter (string)
	#	cutoff: distance cutoff for representation
	#	max: maximum size of molecules

	# Returns:
	#	x: 1D list of features (each item in list is a numpy array)
	#	y: 1D list of NMR parameters
	#	r: 1D list of NMR parameter references (molid, atom(1, atom2))

	# convert flag to target
	target = flag_to_target(targetflag)

	x = []
	y = []
	r = []
	# Loop through molecules to construct features
	for mol in mols:
		# Make representations using qml function
		reps = qml.fchl.generate_representation(mol.xyz, mol.types, max, cut_distance=cutoff)
		# If chemical shift target
		if len(target) == 1:
			for i in range(len(mol.types)):
				if mol.types[i] == target[0]:
					x.append([reps[i]])
					y.append(mol.shift[i])
					r.append([mol.molid, i])
		# If coupling target
		if len(target) == 3:
			for i in range(len(mol.types)):
				for j in range(len(mol.types)):
					if i == j:
						continue

					if not (mol.types[i] == target[1] and mol.types[j] == target[2]):
						continue

					if mol.coupling_len[i][j] != target[0]:
						continue

					if i > j:
						x.append([reps[j], reps[i]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, j, i])
					else:
						x.append([reps[i], reps[j]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, i, j])

	return x, y, r

# Make mbtypes for aSLATM representation
def get_aSLATM_mbtypes(mols):
	# Input:
	#	mols: list of autoENRICH nmrmol objects

	# Returns: mbtypes (list of atom type combinations)

	if len(mols) == 0:
		mbtypes = [[1],[1,1], [1,1,1], [1,1,6], [1,1,7], [1,1,8], [1,1,9], [1,6], [1,6,1], [1,6,6], [1,6,7], [1,6,8], [1,6,9], [1,7], [1,7,1], [1,7,6], [1,7,7], [1,7,8], [1,7,9], [1,8], [1,8,1], [1,8,6], [1,8,7], [1,8,8], [1,8,9], [1,9], [1,9,1], [1,9,6], [1,9,7], [1,9,8], [1,9,9], [6], [6,1], [6,1,1], [6,1,6], [6,1,7], [6,1,8], [6,1,9], [6,6], [6,6,1], [6,6,6], [6,6,7], [6,6,8], [6,6,9], [6,7], [6,7,1], [6,7,6], [6,7,7], [6,7,8], [6,7,9], [6,8], [6,8,1], [6,8,6], [6,8,7], [6,8,8], [6,8,9], [6,9], [6,9,1], [6,9,6], [6,9,7], [6,9,8], [6,9,9], [7],[7,1], [7,1,1], [7,1,6], [7,1,7], [7,1,8], [7,1,9], [7,6], [7,6,1], [7,6,6], [7,6,7], [7,6,8], [7,6,9], [7,7], [7,7,1], [7,7,6], [7,7,7], [7,7,8], [7,7,9], [7,8], [7,8,1], [7,8,6], [7,8,7], [7,8,8], [7,8,9], [7,9], [7,9,1], [7,9,6], [7,9,7], [7,9,8], [7,9,9], [8], [8,1], [8,1,1], [8,1,6], [8,1,7], [8,1,8], [8,1,9], [8,6], [8,6,1], [8,6,6], [8,6,7], [8,6,8], [8,6,9], [8,7], [8,7,1], [8,7,6], [8,7,7], [8,7,8], [8,7,9], [8,8], [8,8,1], [8,8,6], [8,8,7], [8,8,8], [8,8,9], [8,9], [8,9,1], [8,9,6], [8,9,7], [8,9,8], [8,9,9], [9], [9,1], [9,1,1], [9,1,6], [9,1,7], [9,1,8], [9,1,9], [9,6], [9,6,1], [9,6,6], [9,6,7], [9,6,8], [9,6,9], [9,7], [9,7,1], [9,7,6], [9,7,7], [9,7,8], [9,7,9], [9,8], [9,8,1], [9,8,6], [9,8,7], [9,8,8], [9,8,9], [9,9], [9,9,1], [9,9,6], [9,9,7], [9,9,8], [9,9,9]]

	else:
		nuclear_charges = []
		for tmp_mol in mols:
			nuclear_charges.append(tmp_mol.types)
		mbtypes = qml.representations.get_slatm_mbtypes(nuclear_charges)

	return mbtypes

# Make aSLATM features
def get_aSLATM_features(mols, targetflag='CCS', cutoff=5.0, max=400, mbtypes=[]):
	# Input:
	#	mols: list of autoENRICH nmrmol objects
	#	targetflag: flag corresponding to nmr parameter (string)
	#	cutoff: distance cutoff for representation
	#	max: maximum size of molecules
	#	mbtypes: list of all possible atoms, pairs and triplets

	# Returns:
	#	x: 1D list of features (each item in list is a numpy array)
	#	y: 1D list of NMR parameters
	#	r: 1D list of NMR parameter references (molid, atom(1, atom2))

	# convert flag to target
	target = flag_to_target(targetflag)

	x = []
	y = []
	r = []

	# If no mbtypes given, get default ones
	if len(mbtypes) == 0:
		mbtypes = get_aSLATM_mbtypes([])

	# loop through molecules to construct features
	for mol in mols:
		# Basic assertions
		assert mol.xyz.shape[1] == 3
		assert mol.xyz.shape[0] == mol.types.shape[0]
		# Get representations from qml function
		reps = qml.representations.generate_slatm(mol.xyz, mol.types, mbtypes, rcut=cutoff)
		# Convert features to numpy array (no idea why its a list in the first place)
		reps = np.asarray(reps)
		# if chemical shift target
		if len(target) == 1:
			for i in range(len(mol.types)):
				if mol.types[i] == target[0]:
					x.append([reps[i]])
					y.append(mol.shift[i])
					r.append([mol.molid, i])
		# If coupling target
		if len(target) == 3:
			for i in range(len(mol.types)):
				for j in range(len(mol.types)):
					if i == j:
						continue
					if not ( mol.types[i] == target[1] and mol.types[j] == target[2] ):
						continue

					if mol.coupling_len[i][j] != target[0]:
						continue

					if i > j:
						x.append([reps[j], reps[i]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, j, i])
					else:
						x.append([reps[i], reps[j]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, i, j])

	return x, y, r


def get_CMAT_features(mols, targetflag='CCS', cutoff=5.0, max=100,
				central_decay=-1, interaction_cutoff =1e6, interaction_decay=-1):
	# Input:
	#	mols: list of autoENRICH nmrmol objects
	#	targetflag: flag corresponding to nmr parameter (string)
	#	cutoff: distance cutoff for representation
	#	max: maximum size of molecules
	#	central_decay: representation parameter
	#	interaction_cutoff: representation parameter
	#	interaction_decay: representation parameter

	# Returns:
	#	x: 1D list of features (each item in list is a numpy array)
	#	y: 1D list of NMR parameters
	#	r: 1D list of NMR parameter references (molid, atom(1, atom2))

	# Convert flag to target
	target = flag_to_target(targetflag)

	x = []
	y = []
	r = []
	'''
	for mol in mols:
		if len(mol.types) > max:
			max = len(mol.types)
	'''
	# Loop through molecules to make features
	for mol in mols:
		# Get representations from qml function
		reps = qml.representations.generate_atomic_coulomb_matrix(mol.types, mol.xyz, size = max, central_cutoff = cutoff)
		# If chemical shift target
		if len(target) == 1:
			for i in range(len(mol.types)):
				if mol.types[i] == int(target[0]):
					x.append([reps[i]])
					y.append(mol.shift[i])
					r.append([mol.molid, i])
		# If coupling target
		if len(target) == 3:
			for i in range(len(mol.types)):
				for j in range(len(mol.types)):
					if i == j:
						continue

					if not ( mol.types[i] == target[1] and mol.types[j] == target[2] ):
						continue

					if mol.coupling_len[i][j] != target[0]:
						continue

					if i > j:
						x.append([reps[j], reps[i]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, j, i])
					else:
						x.append([reps[i], reps[j]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, i, j])

	return x, y, r

# Get ACSF features
def get_ACSF_features(mols, targetflag='CCS', cutoff=5.0, max=400, elements=[], nRs2=3, nRs3=3, nTs=3, eta2=1,
									eta3=1, zeta=1, acut=5, bin_min=0.8):
	# Input:
	#	mols: list of autoENRICH nmrmol objects
	#	targetflag: flag corresponding to nmr parameter (string)
	#	cutoff: distance cutoff for representation
	#	max: maximum size of molecules
	#	elements: list of possible atom types
	# 	various feature parameters

	# Returns:
	#	x: 1D list of features (each item in list is a numpy array)
	#	y: 1D list of NMR parameters
	#	r: 1D list of NMR parameter references (molid, atom(1, atom2))

	# Convert flag to target
	target = flag_to_target(targetflag)

	x = []
	y = []
	r = []
	# Adjust max size if needed
	for tmp_mol in mols:
		if len(tmp_mol.types) > max:
			max = len(tmp_mol.types)
	# loop through mols to make features
	for mol in mols:
		# Get representations from qml function
		reps = qml.representations.generate_acsf(mol.types, mol.xyz, elements=[1, 6, 7, 8, 9, 14, 15, 16, 17, 35],
												nRs2=int(nRs2), nRs3=int(nRs3),
												nTs=int(nTs), eta2=eta2, eta3=eta3, zeta=zeta, rcut=cutoff, acut=acut,
												bin_min=0.0, gradients=False)
												#pad=1010)
		# If chemical shift target
		if len(target) == 1:
			for i in range(len(mol.types)):
				if mol.types[i] == target[0]:
					x.append([reps[i]])
					y.append(mol.shift[i])
					r.append([mol.molid, i])
		# If coupling target
		if len(target) == 3:
			for i in range(len(mol.types)):
				for j in range(len(mol.types)):
					if i == j:
						continue
					if not ( mol.types[i] == target[1] and mol.types[j] == target[2] ):
						continue

					if mol.coupling_len[i][j] != target[0]:
						continue

					if i > j:
						x.append([reps[j], reps[i]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, j, i])
					else:
						x.append([reps[i], reps[j]])
						y.append(mol.coupling[i][j])
						r.append([mol.molid, i, j])

	return x, y, r



# Argument parser
import argparse

def IMP_parser(args):
	parser = argparse.ArgumentParser(description='IMPRESSION')
	# Define command to perform
	parser.add_argument('Command', help='IMPRESSION command',
							choices=['setup_train', 'train', 'setup_predict', 'predict'])
	# Get preferences
	parser.add_argument('--prefs', help='How to obtain settings for command, leave blank to use command line input, specify json file, or <wizard> to run wizard',
							 default='')
	# Define training set, only needed for train and setup_train commands
	parser.add_argument('--training_set', help='Training dataset file(s), either single csv/pkl file or one/multiple nmredata files',
						 action="store", dest='training_set', default='')
	# Define output directory for non log files
	parser.add_argument('--output_dir', help='Output directory for non log files',
						action="store", dest='output_dir', default='')
	# Check whether to store prepared datasets or not
	parser.add_argument('--store_datasets', help='Option to store datasets as pickle files for later use',
						 action="store", dest='store_datasets', default='')

	# Define list of target parameters
	parser.add_argument('--target_list', help='Optional list of targets to go through',
						 action="store", dest='target_list', default=[])
	# Define single target (largely used internally)
	parser.add_argument('--targetflag', help='target NMR parameter',
							action="store", dest='targetflag', default='')

	# Optional arguments for train/test
	# Type of model to use
	parser.add_argument('--modelflag', help='type of model to use',
							action="store", dest='modelflag', default='',
							choices=['KRR', 'FCHL', 'NN', 'TFM'])
	# Type of features to user
	parser.add_argument('--featureflag', help='type of features to use',
							action="store", dest='featureflag', default='',
							choices=['CMAT', 'aSLATM', 'FCHL', 'ACSF', 'BCAI', ''])

	# Optional argments for train
	# Method for hyper-parameter search
	parser.add_argument('--searchflag', help='Method for hyper-parameter search',
							action="store", dest='searchflag', default='',
							choices=['grid', 'gaussian', 'random', ''])
	# Whether to optimise features in hyper-parameter optimisation (uses defaults for features otherwise)
	# This offers a pretty massive speed improvement especially for some features (aSLATM)
	parser.add_argument('--feature_optimisation', help='HPS includes feature parameters',
							action="store", dest='feature_optimisation', default='',
							choices=['True', 'False', ''])
	# Can supply pre-made features to use
	parser.add_argument('--feature_file', help='File containing pre-made feature dataset object',
							action="store", dest='feature_file', default='')
	# Define ranges for model parameters to optimise
	parser.add_argument('--param_ranges', help='Dictionary of parameter ranges for HPS',
							action="store", dest='param_ranges', default='{}')
	# Specify whether to treat paramater is linear of log scale (helps with optimisation)
	parser.add_argument('--param_logs', help='Dictionary (truth values only) for each parameter specifying whether parameter is on a log scale',
							action="store", dest='param_logs', default='{}')
	# Number of cross validation steps to perform
	parser.add_argument('--cv_steps', help='Number of cross validations to perform',
							 action="store", dest='cv_steps', default=-404)
	# How many HPS iterations to perform
	parser.add_argument('--epochs', help='Number of HPS iterations to perform',
							action="store", dest='epochs', default=-404)
	# Specify output logfile
	parser.add_argument('--logfile', help='Name of output log file',
							action="store", dest='logfile', default='')


	# Optional system arguments
	parser.add_argument('--python_env', help='Name of python environment to be used',
							action="store", dest='python_env', default='')
	# Define system (allows for HPC cluster specific submission)
	parser.add_argument('--system', help='System currently running',
							 action="store", dest='system', default='')
	# How much memory to allow in a HPC submission
	parser.add_argument('--memory', help='Memory needed in submission scripts',
							 action="store", dest='memory', default=-404)
	# How many processors to request in HPC submission
	parser.add_argument('--processors', help='Processors needed in submission scripts',
							 action="store", dest='processors', default=-404)
	# How much walltime to request in HPC submission
	parser.add_argument('--walltime', help='Walltime required for submission',
							 action="store", dest='walltime', default='')

	# Optional argments for predict
	# How many variance models to use
	parser.add_argument('--var', help='Number of pre-prediction variance models',
						action="store", dest='var', default=-404)
	# Define existing models to use for predictions
	parser.add_argument('--models', help='Existing model(s) to use, list',
						 action="store", dest='models', default=[])
	# Current prediction model, not sure if this is used
	parser.add_argument('--model', help='Current prediction model',
						action="store", dest='model', default='none')
	# Define set(s) of molecules to perform predictions on
	parser.add_argument('--test_sets', help='Testing dataset(s), either lists of file search patterns or individual files, list',
						 action="store", dest='test_sets', default=[])


	# Optional argments for grid search
	# Define density of points in grid
	parser.add_argument('--grid_density', help='Point density for grid search',
							 action="store", dest='grid_density', default=-404)

	# Optional arguments for gaussian search
	parser.add_argument('--kappa', help='Kappa value for gaussian process HPS',
							 action="store", dest='kappa', default=-404)
	parser.add_argument('--xi', help='Xi value for gaussian process HPS',
							action="store", dest='xi', default=-404)
	parser.add_argument('--random', help='Frequency of random selection in gaussian process search',
							action="store", dest='random', default=-404)
	parser.add_argument('--load', help='Load previous search from logfile before starting',
							action="store", dest='load', default=False)

	# Optional arguments for make features
	parser.add_argument('--input_files', help='Files to create features from',
							 action="store", dest='input_files', default='none')
	parser.add_argument('--output_files', help='Files to store features in',
							 action="store", dest='output_files', default='none')

	# Be very careful with this, having a different size between the training and
	# test sets screws up the predictions
	parser.add_argument('--max_size', help='Maximum molecule size',
							action="store", dest='max_size', default=-404)
	# Option to trace the code execution, for bug hunting
	parser.add_argument('--tracecode', help='Trace the code execution',
							action="store_true", dest='tracecode', default=False)
	# Option to trace memory usage, prints the biggest 10 memory blocks at the end of execution
	parser.add_argument('--tracemem', help='Trace the memory usage',
								action="store_true", dest='tracemem', default=False)
	# Option to trace code execution by time
	parser.add_argument('--tracetime', help='Trace the execution timings',
								action="store_true", dest='tracetime', default=False)


	return parser.parse_args(args)


def aE_parser(args):
	# Argparser
	parser = argparse.ArgumentParser(description='auto-ENRICH')
	# Define name of molecule (used for save/load of pickle file)
	parser.add_argument('Molecule', help='name of molecule')
	# Define command to perform
	parser.add_argument('Command', help='auto-ENRICH command', choices=['undo', 'init', 'conf_search', 'setup_opt', 'process_opt', 'setup_nmr', 'process_nmr', 'update', 'check_status'])
	# Optional arguments
	parser.add_argument('--xyz', help='xyz file to initialise molecule', action="store", dest='xyz_file', default='None')
	parser.add_argument('--path', help='path to molecule pickle file', action="store", dest='path', default='')
	parser.add_argument('--prefs', help='preferences file to use', action="store", dest='prefs', default='ENRICH.json')

	return parser.parse_args(args)

def util_parser(args):
	# Argparser
	parser = argparse.ArgumentParser(description='aE_utils')

	parser.add_argument('Command', help='aE_utils command',
							choices=['convert_to_nmredata', 'compare'])

	parser.add_argument('--files', help='Files to process',
							action="store", dest='files', default='')

	parser.add_argument('--type', help='File type',
							action="store", dest='type', default='g09')

	parser.add_argument('--out_dir', help='output directory',
							action="store", dest='out_dir', default='')

	parser.add_argument('--comp_targets', help='target parameters to compare',
							action="store", dest='comp_targets', default='HCS')

	parser.add_argument('--comp_sets', help='Datasets to compare',
							action="store", dest='comp_sets', default='')

	parser.add_argument('--comp_labels', help='Labels for datasets',
							action="store", dest="comp_labels", default='1 2')

	parser.add_argument('--match_criteria', help='Criteria by which to match molecules from different sets',
							action="store", dest="match_criteria", choices=['loose', 'strict', 'id'], default='id')

	return parser.parse_args(args)

def combine_args(user_args, file_args, wiz_args, default_args):
	args = {}
	for arg in user_args:
		user = False
		if type(user_args[arg]) is int:
			if user_args[arg] != -404:
				user = True
		elif type(user_args[arg]) is str:
			if user_args[arg] not in ['', 'none', '{}']:
				user = True
		elif type(user_args[arg]) is dict:
			if len(user_args[arg]) != 0:
				user = True
		elif type(user_args[arg]) is list:
			if len(user_args[arg]) != 0:
				user = True

		file = False
		if arg in file_args:
			if type(file_args[arg]) is int:
				if file_args[arg] != -404:
					file = True
			elif type(file_args[arg]) is str:
				if file_args[arg] not in ['', 'none', '{}']:
					file = True
			elif type(file_args[arg]) is dict:
				if len(file_args[arg]) != 0:
					file = True
			elif type(file_args[arg]) is list:
				if len(file_args[arg]) != 0:
					file = True


		if user:
			args[arg] = user_args[arg]
		elif file:
			args[arg] = file_args[arg]
		elif arg in wiz_args:
			args[arg] = wiz_args[arg]
		elif arg in default_args:
			args[arg] = default_args[arg]
		else:
			continue
	return args



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.


from .nmrmol import nmrmol
from autoENRICH.util.file_gettype import get_type
import numpy as np
np.set_printoptions(threshold=99999999999)

class dataset(object):

	def __init__(self):

		self.mols = []
		self.x = []
		self.y = []
		self.r = []
		self.mol_order = []

		self.params = {}


	def get_mols(self, files, type='', label_part=0):
		self.mols = []
		for file in files:
			id = file.split('/')[-1].split('.')[0].split('_')[label_part]
			mol = nmrmol(molid=id)

			if type == '':
				ftype = get_type(file)
			else:
				ftype = type

			mol.read_nmr(file, ftype)
			self.mols.append(mol)
			#print(id)
			#print(mol.coupling_len)


	def get_features_frommols(self, args, params={}):

		featureflag = args['featureflag']
		targetflag = args['targetflag']
		try:
			max = args['max_size']
		except:
			max = 200

		if 'cutoff' in params:
			if params['cutoff'] < 0.1:
				params['cutoff'] = 0.1
		else:
			params['cutoff'] = 5.0


		x = []
		y = []
		r = []

		self.params = params

		if featureflag in ['aSLATM', 'CMAT', 'FCHL', 'ACSF']:
			from autoENRICH.ml.features import QML_features
		elif featureflag in ['BCAI']:
			from autoENRICH.ml.features import TFM_features
		elif featureflag in ['dummy']:
			from autoENRICH.ml.features import GNR_features

		if featureflag == 'dummy':
			for mol in self.mols:
				_x, _y, _r = GNR_features.get_dummy_features([mol], targetflag)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)


		elif featureflag == 'aSLATM':
			mbtypes = QML_features.get_aSLATM_mbtypes(self.mols)
			for mol in self.mols:
				_x, _y, _r = QML_features.get_aSLATM_features([mol], targetflag, params['cutoff'], max=max, mbtypes=mbtypes)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)


		elif featureflag == 'CMAT':

			# Set (not found) parameters to defaults
			if not 'central_decay' in params:
				params['central_decay'] = -1
			if not 'interaction_cutoff' in params:
				params['interaction_cutoff'] = 1000000.0
			if not 'interaction_decay' in params:
				params['interaction_decay'] = -1

			for mol in self.mols:
				if len(mol.types) >max:
					args['max_size'] = len(mol.types)
					print('WARNING, SETTING MAXIMUM MOLECULE SIZE TO, ', max)
			for mol in self.mols:
				_x, _y, _r = QML_features.get_CMAT_features([mol], targetflag, params['cutoff'], args['max_size'], central_decay=params['central_decay'],
														interaction_cutoff=params['interaction_cutoff'], interaction_decay=params['interaction_decay'])
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)


		elif featureflag == 'FCHL':
			for mol in self.mols:
				if len(mol.types) >max:
					max = len(mol.types)
					print('WARNING, SETTING MAXIMUM MOLECULE SIZE TO, ', max)

				'''
				for mol in self.mols:
				_x, _y, _r = QML_features.get_FCHL_features([mol], targetflag, params['cutoff'], max)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)
				'''
			x, y, r = QML_features.get_FCHL_features(self.mols, targetflag, params['cutoff'], max)


		elif featureflag == 'ACSF':
			elements = set()
			for tmp_mol in self.mols:
				elements = elements.union(tmp_mol.types)
			elements = sorted(list(elements))

			# Set (not found) parameters to defaults


			for mol in self.mols:
				_x, _y, _r = QML_features.get_ACSF_features([mol], targetflag, params['cutoff'], elements=elements, nRs2=params['nRs2'],
												nRs3=params['nRs3'], nTs=params['nTs'], eta2=params['eta2'], eta3=params['eta3'], zeta=params['zeta'],
												acut=params['acut'], bin_min=params['bin_min'])
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)

		elif featureflag == 'BCAI':
			x, y, r, mol_order = TFM_features.get_BCAI_features(self.mols, targetflag)

		else:
			print('Feature flag not recognised, no feature flag: ', featureflag)

		if featureflag == 'BCAI':
			self.x = x
			self.y = y
			self.r = r
			self.mol_order = mol_order
		else:
			self.x = np.asarray(x)
			self.y = np.asarray(y)
			self.r = r

		if featureflag not in ['dummy', 'BCAI']:
			print('Reps generated, shape: ', self.x.shape)

	def get_features_fromfiles(self, files, featureflag='CMAT', targetflag='CCS', cutoff=5.0, max=400, mbtypes=[], elements=[]):
		self.params = {}

		x = []
		y = []
		r = []

		if featureflag == 'aSLATM':
			mbtypes = QML_features.get_aSLATM_mbtypes([])
			for file in files:
				id = file.split('.')[0].split('_')[-1]
				mol = nmrmol(id)
				mol.read_nmr(file, 'nmredata')
				_x, _y, _r = QML_features.get_aSLATM_features(mol, targetflag, cutoff, mbtypes)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)


		elif featureflag == 'CMAT':
			for file in files:
				id = file.split('.')[0].split('_')[-1]
				mol = nmrmol(id)
				mol.read_nmr(file, 'nmredata')
				_x, _y, _r = QML_features.get_CMAT_features(mol, targetflag, cutoff, max)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)


		elif featureflag == 'FCHL':
			for file in files:
				id = file.split('.')[0].split('_')[-1]
				mol = nmrmol(id)
				mol.read_nmr(file, 'nmredata')
				_x, _y, _r = QML_features.get_FCHL_features(mol, targetflag, cutoff, max)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)


		elif featureflag == 'ACSF':
			for file in files:
				id = file.split('.')[0].split('_')[-1]
				mol = nmrmol(id)
				mol.read_nmr(file, 'nmredata')
				_x, _y, _r = QML_features.get_ACSF_features(mol, targetflag, cutoff, elements)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)

		elif featureflag == 'BCAI':
			for file in files:
				id = file.split('.')[0].split('_')[-1]
				mol = nmrmol(id)
				mol.read_nmr(file, 'nmredata')
				_x, _y, _r = TFM_features.get_BCAI_features(mol, targetflag, cutoff)
				x.extend(_x)
				y.extend(_y)
				r.extend(_r)

		self.x = np.asarray(x)
		self.y = np.asarray(y)
		self.r = r


	def assign_from_ml(self, pred_y, var, zero=True):
		assert len(self.r) > 0, print('Something went wrong, nothing to assign')

		for mol in self.mols:
			if zero:
				mol.coupling = np.zeros((len(mol.types), len(mol.types)), dtype=np.float64)
				mol.shift = np.zeros((len(mol.types)), dtype=np.float64)

			for r, ref in enumerate(self.r):

				if mol.molid != ref[0]:
					continue

				if len(ref) == 2:
					for t in range(len(mol.types)):
						if ref[1] == t:
							mol.shift[t] = pred_y[r]
							mol.shift_var[t] = var[r]

				elif len(ref) == 3:
					for t1 in range(len(mol.types)):
						for t2 in range(len(mol.types)):
							if ref[1] == t1 and ref[2] == t2:
								mol.coupling[t1][t2] = pred_y[r]
								mol.coupling_var[t1][t2] = var[r]











##



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from autoENRICH.util.flag_handler.hdl_targetflag import flag_to_target
import numpy as np

# Make 'dummy features' for a given target, used to get list of references
def get_dummy_features(mols, targetflag='CCS'):
	# Input:
	#	mols: list of autoENRICH nmrmol objects
	#	targetflag: flag corresponding to nmr parameter (string)

	# Returns:
	#	x: empty list (where features would normally go)
	#	y: 1D list of NMR parameters
	#	r: 1D list of NMR parameter references (molid, atom(1, atom2))

	target = flag_to_target(targetflag)

	x = []
	y = []
	r = []

	for mol in mols:
		if len(target) == 1:
			for i in range(len(mol.types)):
				if mol.types[i] == target[0]:
					y.append(mol.shift[i])
					r.append([mol.molid, i])

		if len(target) == 3:
			for i in range(len(mol.types)):
				for j in range(len(mol.types)):
					if i == j:
						continue

					if not ( mol.types[i] == target[1] and mol.types[j] == target[2] ):
						continue

					if mol.coupling_len[i][j] != target[0]:
						continue

					y.append(mol.coupling[i][j])
					r.append([mol.molid, i, j])

	return x, y, r



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import glob
from autoENRICH.util.flag_handler import hdl_targetflag, flag_combos, paramdict
import os


def get_minimal_args():
	args = {}
	args['training_set'] = ''
	args['output_dir'] = './'
	args['store_datasets'] = True
	args['target_list'] = ['HCS', 'CCS', '1JCH']
	args['targetflag'] = 'HCS'
	args['modelflag'] = 'KRR'
	args['featureflag'] = 'CMAT'

	args['searchflag'] = 'random'
	args['feature_optimisation'] = 'True'
	args['feature_file'] = ''
	args['param_ranges'] = {}
	args['param_logs'] = {}
	args['cv_steps'] = 5
	args['epochs'] = 200
	args['logfile'] = './IMP.log'

	args['python_env'] = 'IMPgen1'
	args['system'] = 'localbox'
	args['memory'] = 2
	args['processors'] = 6
	args['walltime'] = '100:00:00'

	args['var'] = 5
	args['models'] = ['None']
	args['model'] = 'None'
	args['test_sets'] = ['None']

	args['grid_density'] = 5
	args['kappa'] = 5.0
	args['xi'] = 0.2
	args['random'] = -404
	args['load'] = 'false'

	args['input_files'] = ''
	args['output_files'] = ''
	args['max_size'] = 200

	args['tracecode'] = False
	args['tracemem'] = False
	args['tracetime'] = False

	args['prefs'] = 'default'

	return args


# code to run user through selection of input arguments for selected command
# default flag is used to minimise required user input
def run_wizard(args, default=False):

	# wizard section for training commands
	if args['Command'] == 'setup_train' or args['Command'] == 'train':
		# Training set ##############################################################################
		check = False
		# Repeat input loop until succesful
		while not check:
			# Check currently held args (user input or default)
			# If single file specified:
			if args['training_set'].split('.')[-1] == 'pkl' or args['training_set'].split('.')[-1] == 'csv':
				# Attempt to open the file
				try:
					a = open(args['training_set'], 'r')
					check = True
				except Exception as e:
					print(e)
			# Else try and open first file in list of files
			else:
				files = glob.glob(args['training_set'])
				try:
					a = open(files[0], 'r')
					check = True
				except Exception as e:
					print(e)
			# If check remains false here, prompt user for training set input
			if not check:
				args['training_set'] = input("Training set: \n")


		# Store datasets ##############################################################################
		check = False
		while not check:
			# default to storing datasets
			if default:
				args['store_datasets'] = 'True'
				check = True
			# Get user decision y/n
			else:
				decision = input("Do you want to store dataset as pickle after creation ? [y]/n: \n")
				# Go through several reasonable input options
				if len(decision) == 0:
					args['store_datasets'] = 'True'
					check = True
				elif decision[0] in ['n', 'N']:
					args['store_datasets'] = 'False'
					check = True
				elif decision[0] in ['y', 'Y']:
					args['store_datasets'] = 'True'
					check = True
				else:
					check = False

		# Assign default model based on feature selection
		if default:
			if args['featureflag'] == 'FCHL':
				args['modelflag'] = 'FCHL'
			elif args['featureflag'] in ['aSLATM', 'CCS', 'ACSF']:
				args['modelflag'] = 'KRR'
			else:
				args['modelflag'] == 'NN'
		# Alternatively get model from user
		else:
			combocheck = False
			while not combocheck:
				# Model ##############################################################################
				check = False
				while not check:
					model = input("What type of model do you want ? KRR, FCHL, NN, TFM : \n")
					if model in ['KRR', 'FCHL', 'NN', 'TFM']:
						args['modelflag'] = model
						check = True
				# Only one available option for FCHL model, must be FCHL feature
				if model == 'FCHL':
					args['featureflag'] = 'FCHL'
					check = True
					# No need to check combination
					combocheck = True
				# Ask for user input featureflag
				else:
					# Feature ##############################################################################
					check = False
					while not check:
						feature = input("What type of input features do you want ? CMAT, aSLATM, FCHL, ACSF, BCAI : \n")
						if feature in ['CMAT', 'aSLATM', 'FCHL', 'ACSF', 'BCAI']:
							args['featureflag'] = feature
							check = True
						else:
							print('Requested feature [', feature , '] not recognised. . .')

					# Check for model and feature combination, only some actually work
					combocheck = flag_combos.check_combination(args['modelflag'], args['featureflag'])
					if not combocheck:
						print('Invalid combination of model and feature, try again . . .')

		# Target ##############################################################################
		check = False
		while not check:
			if default:
				if args['target_list'] == '':
					args['target_list'] = ['HCS', 'CCS', '1JCH']

			else:
				target_list = input("What target parameter(s) are you interested in ? XCS or nJXY, e.g. HCS CCS 1JCH : \n")
				args['target_list'] = target_list.split()
				print(args['target_list'])

				if type(args['target_list']) != list:
					print(args['target_list'], 'Not a list. . .')
					check = False
					continue

			for target in args['target_list']:
				param = hdl_targetflag.flag_to_target(target)
				if param == 0:
					print('Invalid parameter flag')
					check = False
				else:
					args['targetflag'] = args['target_list'][0]
					check = True

		# Search method ##############################################################################
		check = False
		while not check:
			if default:
				if args['searchflag'] == '':
					args['searchflag'] = 'gaussian'
				check = True
			else:
				searchmethod = input("What search method should be used ? grid, gaussian, random : \n")
				if searchmethod in ['grid', 'gaussian', 'random']:
					args['searchflag'] = searchmethod
					check = True

		# Feature optimisation ##############################################################################
		check = False
		while not check:
			if default:
				args['feature_optimisation'] = 'True'
				check = True

			else:
				feature_opt = input("Do you want to include feature parameters in optimisation ? : [y]/n \n")
				if len(feature_opt) == 0 or feature_opt[0] in ['Y', 'y']:
					args['feature_optimisation'] = 'True'
					check = True
				elif feature_opt[0] in ['N', 'n']:
					args['feature_optimisation'] = 'False'
					check = True

			# Feature file ##############################################################################
			if args['feature_optimisation'] == 'False':
				check = False
				while not check:
					file = input("File containing pre-made features dataset object: \n")
					try:
						a = open(file, 'r')
						check = True
					except Exception as e:
						print(e)

		args['param_ranges'], args['param_logs'] = paramdict.construct_param_dict(args['modelflag'], args['featureflag'], args['targetflag'])

		# Parameters ##############################################################################
		"""
		if not default:
			for param in args['param_ranges'].keys():
				check = False
				IP = input("Optimise {param:<10s} ? (y)/n\n".format(param=param))
				if len(IP) == 0:
					IP = 'y'

				if IP[0] in ['n', 'N']:
					args['param_logs'][param] = 'no'
					check = True

				elif IP[0] in ['y', 'Y']:

					IP = input("Select range for parameter (min, max, log) {param:<10s}: default = {min:<10f}, {max:<10f}, {log:<10s} \n".format(param=param,
																														min=args['param_ranges'][param][0],
																														max=args['param_ranges'][param][1],
																														log=args['param_logs'][param]))
					if len(IP) == 0:
						check = True
					else:
						try:
							range = [float(IP.split(',')[0]), float(IP.split(',')[1])]
							log = IP.split(',')[2]

							args['param_ranges'][param] = range
							args['param_logs'][param] = log

							check = True

						except Exception as e:
							print(e)
		"""


		# grid density ##############################################################################
		check = False
		while not check:

			if default:
				args['cv_steps'] = int(args['cv_steps'])
				check = True
			else:
				try:
					cv = input("Specify number of cross validation iterations: default = {0:<10f} \n".format(args['cv_steps']))
					if len(cv) == 0:
						check = True
					else:
						args['cv_steps'] = int(cv)
						check = True
				except Exception as e:
					print(e)

		if default:
			args['epochs'] = int(args['epochs'])
		else:
			if args['searchflag'] == 'grid':
				# grid density ##############################################################################
				check = False
				while not check:
					try:
						grid = input("Specify grid density for parameters: default = {0:<10f} \n".format(args['grid_density']))
						if len(grid) == 0:
							check = True
						else:
							args['grid_density'] = int(grid)
							check = True
					except Exception as e:
						print(e)

			elif args['searchflag'] in ['random', 'gaussian']:
				# epochs ##############################################################################
				check = False
				while not check:
					try:
						epochs = input("Specify number of epochs to run: default = {0:<10f} \n".format(args['epochs']))
						if len(epochs) == 0:
							check = True
						else:
							args['epochs'] = float(epochs)
							check = True
					except Exception as e:
						print(e)


			# kappa ##############################################################################
			if not default:
				check = False
				while not check:
					try:
						kappa = input("Specify kappa value: default = {0:<10f} \n".format(args['kappa']))
						if len(kappa) == 0:
							check = True
						else:
							args['kappa'] = float(kappa)
							check = True
					except Exception as e:
						print(e)

			# xi ##############################################################################
			if not default:
				check = False
				while not check:
					try:
						xi = input("Specify xi value: default = {0:<10f} \n".format(args['xi']))
						if len(xi) == 0:
							check = True
						else:
							args['xi'] = float(xi)
							check = True
					except Exception as e:
						print(e)
			# random ##############################################################################
			if not default:
				check = False
				while not check:
					try:
						random = input("Specify frequency of random samples: default = {0:<10d} \n".format(args['random']))
						if len(random) == 0:
							check = True
						else:
							args['random'] = int(random)
							check = True
					except Exception as e:
						print(e)

	elif args['Command'] == 'setup_predict' or args['Command'] == 'predict':
		# Model(s) ##############################################################################
		check = False
		while not check:
			try:
				models = input("Specify models to make predictions from: \n")
				args['models'] = models.split()
				if type(args['models']) != list:
					print('not a list')
					continue
				for model in args['models']:
					a = open(model, 'r')
					check = True
			except Exception as e:
				print(e)

		# var model(s) ##############################################################################
		check = False
		while not check:
			var = input("How many models are used for pre-prediction variance ? Default=0\n variance models need to be of the format <model_file_name>_n.pkl\n")
			if len(var) == 0:
				args['var'] = 0
				check = True
			else:
				try:
					args['var'] = int(var)
					check = True
				except Exception as e:
					print(e)

		# input_datasets ##############################################################################
		check = False
		while not check:
			testsets = input("Specify set(s) of molecules to predict\n")
			args['test_sets'] = testsets.split()
			check = True
			for tset in args['test_sets']:
				if '*' in tset:
					try:
						files = glob.glob(tset)
						a = open(files[0], 'r')
					except Exception as e:
						check = False
						print(e)

				else:
					try:
						a = open(tset, 'r')
					except Exception as e:
						check = False
						print(e)

	# output directory ###################################################################
	if not default:
		check = False
		while not check:

			output_dir = input("Set output directory ? default is current directory \n")
			if len(output_dir) == 0:
				output_dir = './'
			elif output_dir[-1] != '/':
				output_dir = output_dir + '/'

			if os.path.isdir(output_dir):
				args['output_dir'] = output_dir
				check = True



	return args






















###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from autoENRICH.reference.periodic_table import Get_periodic_table


# Print molecule properties to csv file
def print_mol_csv(outname, refs, typerefs, values, labels):
	# Input:
	# outname: name of output file
	# refs: molecular property references (m x k)	m = number of properties,
	#												k = number of atoms in reference + 1,
	#												ref[y, z] = (molid, atomid1, atomid2, atomid3, . . . )
	# typerefs: atom types corresponding to the atom ids in refs (m x k)
	#												ref[y, z] = (molid, atomtype1, atomtype2, atomtype3, . . . )
	# values: molecular properties (n, m, k)
	# labels: Labels for molecule sets

	# Returns: None

	# start empty line array (to print at the end)
	lines = []
	# get periodic table array
	p_table = Get_periodic_table()
	# Get number of molecule sets
	sets = len(refs[0])
	# Get first part of header string (molecule set labels for mol IDs)
	idstring = ""
	for x in range(len(refs[0])):
		idstring = idstring + "{label:<s}MOLID,".format(label=labels[x])
	# Get second part of header string (atom references)
	refstring = ""
	for y in range(1, len(refs[0][0])):
		refstring = refstring + "{atom:<s},{type:<s},".format(atom='Atom',
											type='Type',)
	# Get third part of header string (molecule set labels for values)
	valstring = ""
	for z in range(len(values[0])):
		valstring = valstring + "{label:<s}VALUE,".format(label=labels[z])
	# Add header string
	lines.append(idstring+refstring+valstring)
	# Loop through property references
	for i in range(len(refs)):
		# Get molid string
		idstring = ""
		for x in range(len(refs[i])):
			idstring = idstring + "{id:<s},".format(id=refs[i][x][0])
		# Get atomic ref string (atomid, atomtype) x number of atom references
		refstring = ""
		for y in range(1, len(refs[0][0])):
			refstring = refstring + "{atom:<s},{type:<s},".format(atom=str(refs[i][0][y]),
												type=p_table[typerefs[i][0][y-1]])
		# Get value string (value1, value2, . . .)
		valstring = ""
		for z in range(len(values[i])):
			valstring = valstring + "{value:<.4f},".format(value=values[i][z])
		# Construct line and add to list
		lines.append(idstring+refstring+valstring)

	# Print all lines to file
	with open(outname, 'w') as f:
		for line in lines:
			print(line, file=f)



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

def get_unique_part(files):

	parts = []

	for file in files:
		parts.append(file.split('/')[-1].split('.')[0].split('_'))

	label_part = 0

	for p1, part1 in enumerate(parts):
		for p2, part2 in enumerate(parts):
			if p1 == p2:
				continue
			for i in range(len(part1)):
				if part1[i] != part2[i]:
					label_part = i


	return label_part



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import autoENRICH.file_read.orca_read as orcaread
import autoENRICH.file_creation.orca_submission as orcasub
from autoENRICH.file_creation import HPC_submission as HPCsub

def setup_resubmission(molecule, prefs, path=''):

	for conformer in molecule.conformers:
		optstatus = orcaread.get_opt_status(conformer.opt_log)

		if optstatus != 'successful':
			conformer.opt_in = orcasub.make_optin(prefs, conformer, path + 'optimisation/')
			conformer.opt_log = conformer.optin.split('.')[0] + '.log'
			conformer.opt_status = 'pre-submission'
			opt_files.append(conformer.opt_in)

		nmrstatus = orcaread.get_nmr_status(conformer.nmr_log)
		if conformer.nmr_status != 'successful' and conformer.opt_status == 'successful':
			conformer.nmr_in = orcasub.make_nmrin(prefs, conformer, path + 'nmr/')
			conformer.nmr_log = conformer.nmr_in.split('.') + '.log'
			conformer.nmr_status = 'pre-submission'
			nmr_files.append(conformer.nmr_in)

	for tag, in_files in zip(['OPT', 'NMR'], [opt_files, nmr_files]):

		system = prefs['comp']['system']
		if tag == 'OPT':
			memory = prefs['opt']['memory']
			processors = prefs['opt']['processors']
			walltime = prefs['opt']['walltime']
		if tag == 'NMR':
			memory = prefs['nmr']['memory']
			processors = prefs['nmr']['processors']
			walltime = prefs['nmr']['walltime']

		files = len(in_files)
		chunks = HPCsub.get_chunks(files)
		for ck in range(chunks):
			start = (ck * max) + 1
			end = ((ck + 1) * max)
			if end > files:
				end = files

			jobname = 'aE_' + molecule.molid + '_' + str(ck) + tag +  '_RESUB'
			header = HPCsub.make_HPC_header(jobname=jobname, system=system, nodes=1, ppn=processors, walltime=walltime, mem=memory)

			strings = HPCsub.make_orca_batch_submission(prefs, in_files, start, end, ck)

			if prefs['comp']['system'] == 'BC3':
				filename = path + 'RESUB_' + tag + '_' + molecule.molid + '_' + str(ck) + '.qsub'
			elif prefs['comp']['system'] == 'BC4':
				filename = path + 'RESUB_' + tag + '_' + molecule.molid + '_' + str(ck) + '.slurm'
			elif prefs['comp']['system'] == 'localbox':
				filename = path + 'RESUB_' + tag + '_' + molecule.molid + '_' + str(ck) + '.sh'
			with open(filename, 'w') as f:
				for string in header:
					print(string, file=f)
				for string in strings:
					print(string, file=f)

		print('Created ', len(chunks), ' ',  tag, 'resubmission files. . .')


















###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

# make batch / submission scripts for HPC jobs

# Get number of max file chunks to produce job submission scripts for
def get_chunks(files, end=-1, start=-1, max=50):
	# Input:
	# files: number of calculation files
	# end: end file number (-1 reverts to last file)
	# start: start file number (-1 reverts to first file)
	# max: max number of files per submission batch

	# Returns: number of submission scripts required

	# "-1" used as flags to mean use all files
	if start < 0 and end < 0:
		start = 1

		if files > max:
			# Get chunks as remainder
			chunks = files / max
			# Check if any remaining
			if files % max > 0:
				# If so, add additional chunk
				chunks = int(chunks) + 1
		else:
			# Always need at least one chunk
			chunks = 1
			end = files
	# If start andor end are specified
	else:
		# Allow for only one of start/end to be specified
		if start == -1:
			start = 1
		if end == -1:
			end = files
		# If more than 1 chunk needed
		if end - start >= max:
			# Get chunks as remainder
			chunks = files / max
			# Check if any remaining
			if files % max > 0:
				# If so, add additional chunk
				chunks = int(chunks) + 1
		else:
			chunks = 1

	# Ensure chunks is an integer (with division it gets cast as float)
	chunks = int(chunks)

	return chunks

# Make generic part of file for use on HPC
def make_HPC_header(jobname='auto-ENRICH', system='PBS', nodes=1, ppn=1, walltime="100:00:00", mem=3):
	# Input:
	# jobname: name for job on HPC system
	# system: type of system
	# nodes: number of nodes to run on
	# ppn: number of processors per node to run on
	# walltime: time to run for
	# mem: memory to request (in GB)

	# Returns: strings containing lines for file

	strings = []

	# NEED BC4 AND GPU VERSION

	if system == 'PBS':
		strings.append('# submission script for PBS')
		strings.append("#PBS -l nodes={0:<1d}:ppn={1:<1d}".format(nodes, ppn))
		strings.append("#PBS -l walltime={0:<9s}".format(walltime))
		strings.append("#PBS -l mem={0:<1d}GB".format(mem))
		strings.append("#PBD -N {0:<10s}".format(jobname))
		strings.append("cd $PBS_O_WORKDIR")
	elif system == 'slurm':
		# sbatch version
		strings.append('# submission script for slurm')
	elif system == 'localbox':
		strings.append('# submission script for local linux box')

	return strings

def make_HPC_orca_batch_submission(prefs, molname, in_array, start, end, jobname='auto-ENRICH', nodes=1, ppn=1, mem=3, walltime="100:00:00"):
	# Input:
	#	prefs: preferences dictionary
	# 	molname: name of molecule
	#	in_array: file containing names of conformer input files
	#	start: conformer id to start at
	#	end: conformer id to end at
	#	jobname: name for job on cluster
	#	nodes: number of nodes to request
	#	ppn: number of processors per node to run on
	#	mem: memory to request (in GB)
	#	walltime: time to run for

	# Returns: strings containing lines for file

	strings = []
	strings.append('#!/bin/bash')
	if prefs['comp']['system'] == 'PBS':
		strings.append("#PBS -l nodes={0:<1d}:ppn={1:<1d}".format(nodes, ppn))
		strings.append("#PBS -l walltime={0:<9s}".format(walltime))
		strings.append("#PBS -l mem={0:<1d}GB".format(mem))
		if prefs['comp']['parallel']:
			strings.append("#PBS -N {0:>1s}".format(jobname))
			strings.append("#PBS -t {0:>1d}-{1:<1d}".format(start, end))
			strings.append("cd $PBS_O_WORKDIR")
			strings.append("NMRNAME=$(gawk -v y=${{PBS_ARRAYID}} 'NR == y' {0:<5s})".format(in_array))
			strings.append("OUTNAME=$( echo $NMRNAME | sed 's/\.in/\.log/')")
			strings.append("orca ${NMRNAME} > ${OUTNAME}")
		else:
			strings.append("cd $PBS_O_WORKDIR")
			strings.append("for i in $(seq {0:>1d} 1 {1:<1d});".format(start, end))
			strings.append("do")
			strings.append("  NMRNAME=$(gawk -v y=${i} 'NR == y' {0:<5s})".format(in_array))
			strings.append("  base=${NMRNAME::${#NMRNAME}-4}")
			strings.append("  FILE=${base}.log")
			strings.append("OUTNAME=$( echo $NMRNAME | sed 's/.in/.log/')")
			strings.append("  if test -f '$FILE'; then")
			strings.append("    continue")
			strings.append("  else")
			strings.append("    orca ${NMRNAME} > ${OUTNAME}")
			strings.append("  fi")
			strings.append("done")

	elif prefs['comp']['system'] =='BC4':
		print('not done yet . . .')
	elif prefs['comp']['system'] == 'localbox':
		strings.append("NUMBERS=$(seq {0:>1d} {1:<1d})".format(start, end))
		strings.append("for NUM in ${NUMBERS}; do")
		strings.append("  NMRNAME=$(head -n${{NUM}}" + " {0:<5s} | tail -1)".format(in_array))
		strings.append("  OUTNAME=$( echo $NMRNAME | sed 's/.in/.log/')")
		strings.append("  orca ${NMRNAME} > ${OUTNAME}")
		strings.append("done")

	return strings



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from autoENRICH.ml.models.model import genericmodel
from sklearn.model_selection import KFold
import qml
import copy

class FCHLmodel(genericmodel):

	def __init__(self, id='FCHLmodel', x=[], y=[], params={}, model_args={}):
		genericmodel.__init__(self, id, x, y, params, model_args)

		if not 'cutoff' in self.params:
			self.params['cutoff'] = 5.0

		if not 'two_body_scaling' in self.params:
			self.params['two_body_scaling'] = np.sqrt(8)
		if not 'three_body_scaling' in self.params:
			self.params['three_body_scaling'] = 1.6
		if not 'two_body_width' in self.params:
			self.params['two_body_width'] = 0.2
		if not 'three_body_width' in self.params:
			self.params['three_body_width'] = np.pi
		if not 'two_body_power' in self.params:
			self.params['two_body_power'] = 4.0
		if not 'three_body_power' in self.params:
			self.params['three_body_power'] = 2.0
		if not 'cut_start' in self.params:
			self.params['cut_start'] = 1.0
		if not 'alchemy_period_width' in self.params:
			self.params['alchemy_period_width'] = 1.6
		if not 'alchemy_group_width' in self.params:
			self.params['alchemy_group_width'] =1.6


	def train(self,  train_x=[], train_y=[]):
		k = []

		if len(train_x) == 0:
			train_x = self.train_x
		if len(train_y) == 0:
			train_y = self.train_y

		try:
			dimensions = train_x.shape[1]
		except:
			print('training reps not in numpy array (KRRmodel.py, train, l26)')
			train_x = np.asarray(train_x)
			train_y = np.asarray(train_y)

		# reshape x array:
		Xr = []
		for _ in range(dimensions):
			Xr.append([])
		for x in train_x:
			for i in range(dimensions):
				Xr[i].append(x[i])
		Xr = np.asarray(Xr)

		# loop through representation sets in training set
		for d in range(dimensions):
			# append kernel for each representation set
			k.append(qml.fchl.get_atomic_symmetric_kernels(Xr[d], verbose=False,
					two_body_scaling=self.params['two_body_scaling'], three_body_scaling=self.params['three_body_scaling'],
					two_body_width=self.params['two_body_width'], three_body_width=self.params['three_body_width'],
					two_body_power=self.params['two_body_power'], three_body_power=self.params['three_body_power'],
					cut_start=self.params['cut_start'], cut_distance=self.params['cutoff'],
					fourier_order=1, alchemy="periodic-table",
					alchemy_period_width=self.params['alchemy_period_width'], alchemy_group_width=self.params['alchemy_group_width'],
					kernel="gaussian", kernel_args={'sigma':[self.params['sigma']]})[0] + self.params['lamda'] * np.identity(Xr[d].shape[0]))

		# loop through kernels
		K = k[0]
		if len(k) > 1:
			for i in range(1, len(k)):
				# multiply kernels so result is k1 * k2 * k3 ...
				K = K * k[i]

		# get the KRR prefactors, i.e. this is the training of the model
		self.alpha = qml.math.cho_solve(K, train_y)

		# report training state
		self.trained = True


	def predict(self, test_x, train_x=[]):
		ks = []

		if len(train_x) == 0:
			train_x = self.train_x

		assert np.asarray(test_x).shape[1] == np.asarray(train_x).shape[1]

		dimensions = np.asarray(train_x).shape[1]

		# reshape x arrays:
		Xe = []
		Xr = []
		for _ in range(dimensions):
			Xe.append([])
			Xr.append([])
		for x in test_x:
			for i in range(dimensions):
				Xe[i].append(x[i])
		for x in train_x:
			for i in range(dimensions):
				Xr[i].append(x[i])
		Xe = np.asarray(Xe)
		Xr = np.asarray(Xr)

		# loop through representation sets in training set
		for d in range(dimensions):
			# append kernel for each representation set
			ks.append(qml.fchl.get_atomic_kernels(Xe[d], Xr[d], verbose=False,
					two_body_scaling=self.params['two_body_scaling'], three_body_scaling=self.params['three_body_scaling'],
					two_body_width=self.params['two_body_width'], three_body_width=self.params['three_body_width'],
					two_body_power=self.params['two_body_power'], three_body_power=self.params['three_body_power'],
					cut_start=self.params['cut_start'], cut_distance=self.params['cutoff'],
					fourier_order=1, alchemy="periodic-table",
					alchemy_period_width=self.params['alchemy_period_width'], alchemy_group_width=self.params['alchemy_group_width'],
					kernel="gaussian", kernel_args={'sigma':[self.params['sigma']]})[0])


		# loop through kernels
		Ks = ks[0]
		if len(ks) > 1:
			for i in range(1, len(ks)):
				# multiply kernels so result is k1 * k2 * k3 ...
				Ks = Ks * ks[i]
		Ks = np.asarray(Ks)

		# predict values of y
		y_pred = np.dot(Ks, self.alpha)

		return y_pred




















##



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from autoENRICH.molecule.dataset import dataset
from autoENRICH.util.flag_handler.hdl_targetflag import flag_to_target
from autoENRICH.util.filename_utils import get_unique_part
from autoENRICH.analysis.compare_mols import mol_isequal
from autoENRICH.file_creation.data_formats.csv import print_mol_csv
import glob
import numpy as np

def compare_datasets(args):

	att_mols = []
	sets = []
	for set_list in args['comp_sets']:
		print('Getting molecules from ', set_list)
		set = dataset()
		label_part = get_unique_part(glob.glob(set_list))
		set.get_mols(glob.glob(set_list), label_part=label_part)
		print(len(set.mols), ' molecules found from ', len(glob.glob(set_list)), ' files')

		sets.append(set)

	assert len(sets) > 1, print('Only one set found. . .')
	assert len(sets[0].mols) == len(sets[1].mols), print('Different numbers of molecules in sets')

	found = []
	for m1, mol1 in enumerate(sets[0].mols):
		if m1 in found:
			continue
		for m2, mol2 in enumerate(sets[1].mols):

			if args['match_criteria'] == 'id':
				if mol1.molid == mol2.molid:
					att_mols.append([mol1, mol2])
				else:
					continue


			if not mol_isequal(mol1, mol2):
				continue

			if [mol1, mol2] in att_mols:
				continue

			if len(sets) > 2:
				for m3, mol3 in enumerate(sets[2].mols):
					if not mol_isequal(mol1, mol3):
						continue

					if [mol1, mol2, mol3] in att_mols:
						continue

					att_mols.append([mol1, mol2, mol3])

			else:
				found.append(m1)
				att_mols.append([mol1, mol2])

	print(len(att_mols), ' molecules matched, out of ', len(sets[0].mols))



	for targetflag in args['comp_targets']:
		print(targetflag)
		target = flag_to_target(targetflag)

		for set in sets:
			set.get_features_frommols({'featureflag': 'dummy',
										'targetflag': targetflag,
										'max_size': 0})

		values = []
		refs = []
		typerefs = []

		assert len(sets[0].r) == len(sets[1].r)
		if len(sets) > 2:
			assert len(sets[2].r) == len(sets[1].r)

		for group in att_mols:
			for i in range(len(sets[0].r)):
				ref1 = sets[0].r[i]
				if ref1[0] != group[0].molid:
					continue
				val1 = sets[0].y[i]
				typeref1 = [group[0].types[row] for row in ref1[1:]]

				for j in range(len(sets[0].r)):

					ref2 = sets[1].r[j]

					if ref2[0] != group[1].molid:
						continue
					val2 = sets[1].y[j]
					typeref2 = [group[1].types[row] for row in ref2[1:]]

					bad = False
					for xx in range(1, len(ref1)):
						if ref1[xx] != ref2[xx]:
							bad = True
					if typeref1 != typeref2:
						#print(typeref1, typeref2)
						bad = True
					if bad:
						continue

					if len(sets) > 2:
						for k in range(len(sets[0].r)):
							ref3 = sets[2].r[k]
							if ref3[0] != group[2].molid:
								continue
							val3 = sets[2].y[k]
							typeref3 = [group[2].types[row] for row in ref3[1:]]


							bad = False
							for xx in range(1, len(ref1)):
								if ref1[xx] != ref3[xx]:
									bad = True
							if bad:
								continue

							refs.append([ref1, ref2, ref3])
							values.append([val1, val2, val3])
							typerefs.append([typeref1, typeref2, typeref3])
					else:
						refs.append([ref1, ref2])
						values.append([val1, val2])
						typerefs.append([typeref1, typeref2])

		outname = 'Comparison_' + str(targetflag) + '.csv'
		print_mol_csv(outname, refs, typerefs, values, args['comp_labels'])

		x = [row[0] for row in values]
		y = [row[1] for row in values]

		MAE = np.mean(np.absolute(np.asarray(x)-np.asarray(y)))
		MAEstring = '{0:<6.3f}'.format(MAE)
		print('MAE between ', args['comp_labels'][0], args['comp_labels'][1], ' = ', MAEstring, '   no. of envs. ', len(x))
















##



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from sklearn.model_selection import KFold
import pickle
import numpy as np
import copy
import time

def setup_logfile(args):
	strings = []
	strings.append('HPS ' + args['searchflag'] + ' SEARCH')
	strings.append(args['modelflag'] + '   ' + args['featureflag'] + '   ' + args['targetflag'])
	for param in args['param_list']:
		if args['param_logs'][param] == 'no':
			continue
		strings.append('{param:<10s}: {low:>10.4g}  <--->  {high:<10.4g}'.format(param=param,
																	low=args['param_ranges'][param][0],
																	high=args['param_ranges'][param][1]))
	strings.append('')
	strings.append('START')
	string = '{i:<10s}\t{score:<10s}'.format(i='i', score='SCORE')
	for param in args['param_list']:
		if args['param_logs'][param] == 'no':
			continue
		string = string + '\t{param:<20s}'.format(param=param)
	string = string + '\t{time:<10s}'.format(time='Mins')
	strings.append(string)

	if args['logfile'] == "":
		args['logfile'] = args['modelflag'] + '_' + args['featureflag'] + '_' + args['targetflag'] + args['searchflag'] + '.log'

	with open(args['logfile'], 'w') as f:
		for string in strings:
			print(string, file=f)

def HPS_iteration(iter, dataset, args, next_point_to_probe={}, BEST_SCORE=100000.00, BEST_PARAMS={}):

	print('HPS_iteration. . .')
	time0 = time.time()

	args['max_size'] = 200
	if args['featureflag'] == 'BCAI' and iter == 0:
		print('Making representations')
		dataset.get_features_frommols(args, params=next_point_to_probe)
	elif args['featureflag'] != 'BCAI' and args['feature_optimisation'] == 'True':
		dataset.get_features_frommols(args, params=next_point_to_probe)

	print('Number of molecules in training set: ', len(dataset.mols))
	print('Number of envs in training set: ', len(dataset.r))

	assert len(dataset.x) > 0
	assert len(dataset.y) > 0

	# create model
	# yes i know this is super bad "practice"
	# if you can think of a better way of allowing QML code to not be screwed up by TF or PyTorch, etc then let me know
	if args['modelflag'] == 'KRR':
		from autoENRICH.ml.models import KRRmodel
		model = KRRmodel.KRRmodel(id, dataset.x, dataset.y, params=next_point_to_probe, model_args=args)
	elif args['modelflag'] == 'FCHL':
		from autoENRICH.ml.models import FCHLmodel
		model = FCHLmodel.FCHLmodel(id, dataset.x, dataset.y, params=next_point_to_probe, model_args=args)
	elif args['modelflag'] == 'NN':
		from autoENRICH.ml.models import NNmodel
		model = NNmodel.NNmodel(id, dataset.x, dataset.y, params=next_point_to_probe, model_args=args)

	if args['cv_steps'] == 1:
		score = model.cv_predict(args['cv_steps'])
	else:
		y_pred = model.cv_predict(args['cv_steps'])
		score = np.mean(np.absolute(y_pred - dataset.y))

	if args['cv_steps'] != 1:
		score = np.mean(np.absolute(y_pred - dataset.y))
	else:
		score = y_pred

	if score > 99999.99 or np.isnan(score):
		score = 99999.99


	time1 = time.time()

	with open(args['logfile'], 'a') as f:
		string = '{i:<10d}\t{score:<10.5f}'.format(i=iter, score=score)
		for param in args['param_list']:
			if args['param_logs'][param] == 'no':
				continue
			string = string + '\t{param:<20.4g}'.format(param=next_point_to_probe[param])

		string = string + '\t{time:<10.4f}'.format(time=(time1-time0)/60)
		print(string, file=f)

	if score < BEST_SCORE:
		BEST_SCORE = score
		BEST_PARAMS = next_point_to_probe
		print('BEST_SCORE = ', score)
	else:
		print('score = ', score)

	if BEST_PARAMS == {}:
		BEST_PARAMS = next_point_to_probe

	return score, BEST_SCORE, BEST_PARAMS

def save_models(dataset, BEST_PARAMS, args):

	# create model
	if args['modelflag'] == 'KRR':
		from autoENRICH.ml.models import KRRmodel
		model = KRRmodel.KRRmodel(id, np.asarray(dataset.x), np.asarray(dataset.y), params=BEST_PARAMS, model_args=args)
	elif args['modelflag'] == 'FCHL':
		from autoENRICH.ml.models import FCHLmodel
		model = FCHLmodel.FCHLmodel(id, np.asarray(dataset.x), np.asarray(dataset.y), params=BEST_PARAMS, model_args=args)
	elif args['modelflag'] == 'NN':
		from autoENRICH.ml.models import NNmodel
		model = NNmodel.NNmodel(id, dataset.x, dataset.y, params=BEST_PARAMS, model_args=args)

	model.train()

	outname = args['output_dir'] +  args['modelflag'] + '_' + args['targetflag'] + '_' + args['featureflag'] + '_' + args['searchflag'] + '_model.pkl'

	pickle.dump(model, open(outname, "wb"))

	kf = KFold(n_splits=args['cv_steps'])
	kf.get_n_splits(dataset.x)

	i = 0
	for train_index, _ in kf.split(dataset.x):
		i += 1

		tmp_model = copy.deepcopy(model)

		tmp_dataset.x = np.asarray(dataset.x)[train_index]
		tmp_model.train_y = np.asarray(model.train_y)[train_index]
		tmp_model.train()

		outfile = args['output_dir'] + outname.split('/')[-1].split('.')[0] + '_' + str(i) + '.pkl'
		pickle.dump(tmp_model, open(outfile, "wb"))

	return outname












#



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import sys
from autoENRICH.file_read import orca_read, g09_read, nmredata_read, structure_read
from autoENRICH.reference.periodic_table import Get_periodic_table
import pickle


class nmrmol(object):
	"""
		nmr data file
	"""

	def __init__(self, molid, path=''):
		self.path = path

		self.molid = str(molid)
		self.types = []
		self.xyz = []
		self.conn = []

		self.shift = []
		self.shift_var = []

		self.coupling = []
		self.coupling_var = []
		self.coupling_len = []

		self.energy = -404.404


	def read_structure(self, file, type):
		old_type_array = self.types
		old_xyz_array = self.xyz
		self.xyz, self.types, self.conn, self.coupling_len = structure_read.generic_pybel_read(file, type)

		if not (np.array_equal(old_type_array, self.types) or np.array_equal(old_xyz_array, self.xyz)):
			self.shift = []
			self.shift_var = []
			self.coupling = []
			self.coupling_var = []

	def read_opt(self, file, type):
		if type == 'orca':
			#self.xyz, self.types, self.conn, self.coupling_len = orca_read.read_structure(file)
			self.energy = orca_read.read_opt(file)


	def read_nmr(self, file, type):
		if type == 'orca':
			self.xyz, self.types, self.conn, self.coupling_len = orca_read.read_structure(file)
			self.shift, self.coupling = orca_read.read_nmr(file)
		elif type == 'g09':
			self.xyz, self.types, self.conn, self.coupling_len = structure_read.generic_pybel_read(file, 'g09')
			self.shift, self.coupling = g09_read.read_nmr(file, len(self.types))
			self.shift_var = np.zeros((len(self.types)), dtype=np.float64)
			self.coupling_var = np.zeros((len(self.types), len(self.types)), dtype=np.float64)
		elif type == 'nmredata':
			self.xyz, self.types, self.conn, _ = structure_read.fast_generic_pybel_read(file, 'sdf')
			self.shift, self.shift_var, self.coupling, self.coupling_var, self.coupling_len = nmredata_read.read_nmr(file, len(self.types))
		else:
			self.xyz, self.types, self.conn, self.coupling_len = structure_read.generic_pybel_read(file, type)
			atoms = len(self.types)
			self.shift = np.zeros((atoms), dtype=np.float64)
			self.shift_var = np.zeros((atoms), dtype=np.float64)
			self.coupling = np.zeros((atoms, atoms), dtype=np.float64)
			self.coupling_var = np.zeros((atoms, atoms), dtype=np.float64)
			print('non-nmr file detected, generated dummy file'.format(type))

	def scale_shifts(self, scaling_factors={}):
		periodic_table = Get_periodic_table()
		for nucleus, factor in scaling_factors.items():
			if nucleus in ['basis_set', 'functional']:
				continue

			for i in range(len(self.shift)):
				if periodic_table[self.types[i]] == nucleus:
					self.shift[i] = (self.shift[i] - factor[1]) / float(factor[0])

	def save_pickle(file):
		pickle.dump(self, open(file, "wb"))














####



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

# Simple get yes or no from user
def yes_or_no():
	reply = str(input(' (y/n): ')).lower().strip()
	if len(reply) == 0:
		return True
	if reply[0] == 'y':
		return True
	if reply[0] == 'n':
		return False
	else:
		return False



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import bayes_opt as bayes
from bayes_opt import BayesianOptimization
from bayes_opt import UtilityFunction
from bayes_opt.observer import JSONLogger
from bayes_opt.event import Events
from bayes_opt.util import load_logs

import pickle
import numpy as np

import copy

from autoENRICH.ml.hyperparameter_tuning import HPS_generic as generic

# Load iterations from previous run
def load_logs(args):
	# Input:
	# args: global argument dictionary

	# Returns: list of paremeters and scores to read into gaussian process
	to_load = []
	try:
		with open(args['logfile'], 'r') as f:
			switch = False
			for line in f:
				if 'START' in line:
					switch = True
				elif switch:
					items = line.split()
					if items[0] == 'i':
						headers = items[1:]
						headers = [x.strip(' ') for x in headers]

					else:
						params = {}
						for i in range(2, len(headers)):
							if headers[i-1] == 'Mins':
								continue
							params[headers[i-1]] = float(items[i])
						score = float(items[1])

						to_load.append([params, score])
	except Exception as e:
		print(e)
		return []

	return to_load


def gaussian_search(dataset, args):

	# determine whether log dictionary was provided
	if len(args['param_logs']) == 0:
		check_logs = False
	else:
		check_logs = True

	pbounds = {}
	for param in args['param_ranges'].keys():
		if args['param_logs'][param] == 'no':
			continue
		#pbounds[param] = (args['param_ranges'][param][0], args['param_ranges'][param][1])
		pbounds[param] = (0, 1)

	optimizer = BayesianOptimization(
		f=None,
		pbounds=pbounds,
		verbose=0, # verbose = 1 prints only when a maximum is observed, verbose = 0 is silent
		random_state=None
	)
	utility = UtilityFunction(kind="ucb", kappa=args['kappa'], xi=args['xi'])

	if args['load']:
		print('Loading previous data. . .')
		to_load = load_logs(args)
		for log in to_load:
			optimizer.register(params=log[0], target=(99999.99-log[1])/99999.99)
		print('Loaded ', len(to_load), ' datapoints')

	generic.setup_logfile(args)

	if args['load']:
		for params, score in to_load:
			with open(args['logfile'], 'a') as f:
				iter = -1
				string = '{i:<10d}\t{score:<10.5f}'.format(i=iter, score=score)
				for param in args['param_list']:
					if args['param_logs'][param] == 'no':
						continue
					string = string + '\t{param:<20.4g}'.format(param=params[param])
				print(string, file=f)

	BEST_SCORE = 999.999
	BEST_PARAMS = {}
	searched = []
	force_random = False
	for e in range(int(args['epochs'])):

		if force_random or (args['random'] > 0 and e%args['random'] == 0):
			print('RANDOM')
			next_point_to_probe = {}
			for param in args['param_ranges'].keys():
				if args['param_logs'][param] == 'no':
					continue

				next_point_to_probe[param] = np.random.uniform(0, 1)
				#next_point_to_probe[param] = np.random.uniform(args['param_ranges'][param][0], args['param_ranges'][param][1])
			force_random = False
		else:
			next_point_to_probe = optimizer.suggest(utility)


		for param in args['param_ranges'].keys():
			if args['param_logs'][param] == 'no':
				continue
			next_point_to_probe[param] = (next_point_to_probe[param]*(args['param_ranges'][param][1]-args['param_ranges'][param][0])) + args['param_ranges'][param][0]

		if check_logs:
			for param in args['param_ranges'].keys():
				if args['param_logs'][param] == 'no':
					continue
				if 'log' in args['param_logs'][param]:
					next_point_to_probe[param] = 10**next_point_to_probe[param]

		score, BEST_SCORE, BEST_PARAMS = generic.HPS_iteration(e, dataset, args, next_point_to_probe=copy.copy(next_point_to_probe),
															BEST_SCORE=BEST_SCORE, BEST_PARAMS=BEST_PARAMS)

		searched.append(next_point_to_probe)
		try:
			optimizer.register(params=next_point_to_probe, target=(99999.99-score)/99999.99)
		except KeyError as e:
			print(e)
			print('Non-unique point generated, (increasing kappa value and) skipping (kappa = ',args['kappa'],')')
			args['kappa'] += 1.0
			utility = UtilityFunction(kind="ei", kappa=args['kappa'], xi=args['xi'])
			print('(forcing random point)')
			force_random = True


	outname = generic.save_models(dataset, BEST_PARAMS, args)
	print('Optimised model(s) saved in ', outname)

	return dataset, BEST_SCORE

























###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from .user_util import yes_or_no
from autoENRICH.molecule.molecule import molecule as moleculeclass
import glob
import os

def check_proceed(stage, command, molecule, path):
	proceed = False

	if command == 'init':
		proceed = True

	if command == 'undo':
		proceed = True

	elif command == 'conf_search':
		if stage != 'init':
			print('Conformational search already run for molecule,', molecule.molid, '. . . continuing will erase existing molecule data.')
			print('Do you want to continue ?')
			# Get user conformation for overwrite
			answer = yes_or_no()
			if answer:
				# enumerate files in conf_search directory then delete them
				files_to_remove = glob.glob(path + 'conf_search/*')
				for file in files_to_remove:
					try:
						os.remove(file)
					except:
						continue
				# remove conf_search directory
				try:
					os.rmdir(path + 'conf_search')
				except:
					pass
				# Re-initialise molecule from existing molecule, retaining only xyz and types
				init_xyz = molecule.xyz
				init_types = molecule.types
				molecule = moleculeclass(init_xyz, init_types, name=molecule.molid, path=path)
				proceed = True
			else:
				proceed = False
		else:
			proceed = True

	elif command == 'setup_opt':
		if stage in ['init', 'pre-opt']:
			proceed = True
		else:
			print('Already setup optimisations, do you want to continue ? This may overwrite/remove existing com/log files')
			answer = yes_or_no()
			if answer:
				molecule.stage='pre-opt'
				proceed = True
			else:
				proceed = False

	elif command == 'process_opt':
		if stage in ['running-opt']:
			proceed = True
		if stage in ['init', 'pre-opt']:
			proceed = False
			print('Optimisation not set up yet, nothing to process. . .')
		else:
			proceed = True

	elif command == 'setup_nmr':
		if stage in ['running-opt', 'pre-nmr']:
			proceed = True
		elif stage in ['init', 'pre-opt']:
			print('Not ready for NMR calculations, need to run init, conf_search, and setup_opt first')
			proceed = False
		else:
			print('Already setup NMR calculations, do you want to continue ? This may overwrite/remove existing com/log files')
			answer = yes_or_no()
			if answer:
				proceed = True
			else:
				proceed = False

	elif command == 'process_nmr':
		if stage in ['running-nmr']:
			proceed = True
		elif stage in ['init', 'pre-opt', 'running-opt']:
			print('Not ready to process NMR calculations, need to run init, conf_search, setup_opt, and setup_nmr first')
			proceed = False
		elif stage in ['post']:
			print('Re-running NMR processing')
			proceed = True

	elif command == 'update':
		if stage in ['init', 'pre-opt']:
			print('Not ready to update, must be running gaussian calculations, need to run init, conf_search and setup_opt first')
			proceed = False
		else:
			proceed = True

	elif command == 'check_status':
		proceed = True


	return proceed, molecule












##



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import glob
from .nmrmol import nmrmol

class conformer(nmrmol):

	def __init__(self, molid, path=''):
		nmrmol.__init__(self, path=path, molid=molid)

		# store location of conf search xyz file
		self.xyz_file = 'None'

		# store location and status of optimisation files
		self.opt_in = 'None'
		self.opt_log = 'None'
		self.opt_status = 'None'

		# store location and status of NMR files
		self.nmr_in = 'None'
		self.nmr_log = 'None'
		self.nmr_status = 'None'

		self.energy = 404.404
		self.pop = 404.404




###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import pybel as pyb
from autoENRICH.pybel_helpers import pybel_bonds, pybel_read
import numpy as np

# Read structure from file using pybel
def generic_pybel_read(file, type):
	# Input:
	#	file: filename
	#	type: type of file (xyz, g09, mol2, . . .)

	# Returns:
	#	xyz: xyz array for structure (2D numpy array)
	#	types: array of atom types (1D numpy array)
	#	conn_table: array of atom connectivity (bond types) (2D numpy array)
	#	coupling_len: array of atom connection distances (2D numpy array)

	mol = next(pyb.readfile(type, file))
	type_list, types = pybel_read.mol_read_type(mol)
	xyz = pybel_read.mol_read_xyz(mol)

	conn_table = pybel_bonds.mol_get_bond_table(mol)

	coupling_len = pybel_bonds.get_coupling_lengths(mol, types, maxlen=6)

	return xyz, types, conn_table, coupling_len

# Read structure from file, but skip finding coupling distances (computationally expensive)
def fast_generic_pybel_read(file, type):
	# Input:
	#	file: filename
	#	type: type of file (xyz, g09, mol2, . . .)

	# Returns:
	#	xyz: xyz array for structure (2D numpy array)
	#	types: array of atom types (1D numpy array)
	#	conn_table: array of atom connectivity (bond types) (2D numpy array)
	#	coupling_len: array of atom connection distances (2D numpy array of zeros)

	mol = next(pyb.readfile(type, file))
	type_list, types = pybel_read.mol_read_type(mol)
	xyz = pybel_read.mol_read_xyz(mol)

	conn_table = pybel_bonds.mol_get_bond_table(mol)

	#coupling_len = pybel_bonds.get_coupling_lengths(mol, types, maxlen=6)
	coupling_len = np.zeros((len(types), len(types)), dtype=np.float64)

	return xyz, types, conn_table, coupling_len



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import json
import sys


def read_prefs(file):
	try:
		with open(file, 'r') as json_file:
			prefs = json.load(json_file)
	except Exception as E:
		print('Error reading preferences file ', file)
		print(E)
		print('Exiting. . .')
		sys.exit(0)


	return prefs

def write_default_prefs(file):

	prefs = {}
	prefs['conf_search'] = {}
	prefs['conf_search']['iterations'] = 200
	prefs['conf_search']['maxconfs'] = 100
	prefs['conf_search']['RMSthresh'] = 0.1
	prefs['conf_search']['Ethresh'] = 100000

	prefs['mol'] = {}
	prefs['mol']['charge'] = 0
	prefs['mol']['multiplicity'] = 1

	prefs['comp'] = {}
	prefs['comp']['parallel'] = True
	prefs['comp']['system'] = 'BC3'
	prefs['comp']['python_env'] = 'env_IMP'
	prefs['comp']['aE_directory'] = "../../aE/"

	prefs['optimisation'] = {}
	prefs['optimisation']['memory'] = 12
	prefs['optimisation']['processors'] = 4
	prefs['optimisation']['opt'] = 'tight'
	prefs['optimisation']['freq'] = False
	prefs['optimisation']['functional'] = 'mPW1PW'
	prefs['optimisation']['basisset'] = '6-311g(d,p)'
	prefs['optimisation']['solvent'] = 'none'
	prefs['optimisation']['grid'] = 'ultrafine'
	prefs['optimisation']['custom_cmd_line'] = False
	prefs['optimisation']['nodes'] = 1
	prefs['optimisation']['walltime'] = '100:00:00'

	prefs['NMR'] = {}
	prefs['NMR']['memory'] = 12
	prefs['NMR']['processors'] = 4
	prefs['NMR']['functional'] = 'wB97X-D3'
	prefs['NMR']['basisset'] = '6-311g(d,p)'
	prefs['NMR']['aux_basis_set'] = 'def2/JK'
	prefs['NMR']['solvent'] = 'none'
	prefs['NMR']['custom_cmd_line'] = False
	prefs['NMR']['nodes'] = 1
	prefs['NMR']['walltime'] = '100:00:00'
	prefs['NMR']['shift_nuclei'] = ['H', 'C', 'N', 'O', 'F']
	prefs['NMR']['spin_nuclei'] = ['H', 'C']
	prefs['NMR']['spin_thresh'] = 20.0

	json.dump(prefs, open(file, 'w'), indent=4)



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openbabel
from autoENRICH.reference.periodic_table import Get_periodic_table

def mol_find_all_paths(mol, start, end, coupling_length, path=[]):
		# append atom to start
		path = path + [start]
		# check if we have reached target atom
		if start == end:
			# if we have, return succesful path
			return [path]
		# define new path
		paths = []
		# loop over neighbouring atoms
		for nbr_atom in openbabel.OBAtomAtomIter(mol.atoms[start].OBAtom):
			# get ID of neighbour
			node = nbr_atom.GetId()
			# check the neighbour is not already in the path, and that the path is not over the required length
			if node not in path and len(path) <= coupling_length:
				# get new paths for the neighbour
				newpaths = mol_find_all_paths(mol, node, end, coupling_length, path)
				#for each new path, check for paths of correct length
				for newpath in newpaths:
					if len(newpath) == coupling_length+1:
						paths.append(newpath)
		return paths

def mol_get_bond_table(mol):

	atoms = len(mol.atoms)

	bond_table = np.zeros((atoms, atoms), dtype=np.int32)

	for atom1 in range(atoms):
		for atom2 in range(atom1, atoms):

			for nbr_atom in openbabel.OBAtomAtomIter(mol.atoms[atom1].OBAtom):
				check = nbr_atom.GetId()
				if atom2 != check:
					continue

				bond = mol.atoms[atom1].OBAtom.GetBond(nbr_atom)
				order = bond.GetBondOrder()

				bond_table[atom1][atom2] = int(order)
				bond_table[atom2][atom1] = int(order)

	return bond_table

def get_coupling_lengths(mol, types, maxlen=6):
	coupling_len = np.zeros((len(types), len(types)), dtype=np.int32)
	for atom1 in range(len(types)):
		for atom2 in range(len(types)):
			coupling_paths = []
			for i in range(maxlen):
				coupling_paths.extend(mol_find_all_paths(mol, atom1, atom2, i))
			length = 999
			for path in coupling_paths:
				if len(path) < length and len(path) != 0:
					length = len(path) - 1

			if length > maxlen+1:
				length = 0

			coupling_len[atom1][atom2] = length
			coupling_len[atom2][atom1] = length

	return coupling_len



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

# Get status of gaussian optimisation file
def get_opt_status(file):
	# Input:
	#	file: filename

	# Returns: status (string), successful, failed, or not submitted

	status = 'unknown'
	try:
		# Read through file looking for SUCCESS or ERROR
		with open(filename, 'r') as f:
			for line in f:
				if 'SUCCESS' in line:
					status = 'successful'
				if 'ERROR' in line:
					status = 'failed'
	except:
		status = 'not submitted'

	return status

# Get status of gaussian nmr file
def get_nmr_status(file):
		# Input:
		#	file: filename

		# Returns: status (string), successful, failed, or not submitted
	status = 'unknown'
	with open(filename, 'r') as f:
		# Read through file looking for SUCCESS or ERROR
		for line in f:
			if 'SUCCESS' in line:
				status = 'successful'
			if 'ERROR' in line:
				status = 'failed'
	return status

# Get energy from optimisation file
def read_opt(file):
	# Input:
	#	file: filename

	# Returns: energy

	# NEEDS TO BE WRITTEN

	return energy

# Get functional and basis set from gaussian log file
def read_functional(file):
	# Input:
	# file: filename

	# Returns: functional, basisset

	functional = ''
	basisset = ''
	with open(file, 'r') as f:
		for line in f:
			# assumes nmrcommand of format:
			# #T nmr(stuff)functional/basisset stuff...
			if '#T nmr' in line:
				try:
					items = line.split()
					nmrcommand = items[2]
					functional = nmrcommand.split('/')[0]
					basisset = nmrcommand.split('/')[1]

					break
				except Exception as e:
					print(line, items)
					print(e)

	return functional, basisset

# Get NMR parameters from gaussian log file
def read_nmr(file, atomnumber):
	# Input:
	#	file: filename
	#	atomnumber: number of atoms in molecule

	# Returns: shift_array (1D numpy array), couplings (2D numpy array)

	# Define empty array for shifts
	shift_array = np.zeros(atomnumber, dtype=np.float64)
	switch = False
	# Go through file to find magnetic shielding tensors
	with open(file, 'r') as f_handle:
		for line in f_handle:
			# If tensor label is found, activate switch
			if "SCF GIAO Magnetic shielding tensor (ppm)" in line:
				switch = True
			# This label comes at the end of the tensor section, so deactivate switch
			if "Fermi Contact" in line:
				switch = False

			if switch:
				# Find isotropic tensor lines
				if "Isotropic" in line:
					items = line.split()
					try:
						num = int(items[0])
					except:
						continue
					# Isotropic shielding tensor is the 5th item (0, 1, 2, 3, '4')
					shift_array[num-1] = float(items[4])

	# Define empty array for couplings
	couplings = np.zeros((atomnumber, atomnumber), dtype=float)
	# Go through file to find coupling constants
	with open(file, 'r') as f:
		switch = False
		for line in f:
			# If coupling label is found, activate switch
			if "Total nuclear spin-spin coupling J (Hz):" in line:
				switch = True
				continue
			# This label comes at the end of the coupling section, so deactivate switch
			elif "End of Minotr" in line:
				switch = False
				continue

			if switch:
				# All coupling lines contain "D", all index lines do not
				if "D" not in line:
					# Get indices for this section
					tokens = line.split()
					i_indices = np.asarray(tokens, dtype=int)
				else:
					# Assign couplings (array is diagonalised in log file, so this is fiddly)
					tokens = line.split()
					index_j = int(tokens[0]) - 1
					for i in range(len(tokens)-1):
						index_i = i_indices[i] - 1
						coupling = float(tokens[i+1].replace("D","E"))
						couplings[index_i][index_j] = coupling
						couplings[index_j][index_i] = coupling

	return shift_array, couplings



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.


import numpy as np

# Read NMR data from nmredata file
def read_nmr(file, atoms):
	# Input:
	#	file: filename
	#	atoms: number of atoms in molecule (read sdf part first)

	# Returns:
	#	shift_array, array of chemical shifts (1D numpy array)
	#	shift_var, array of chemical shift variances (used in machine learning) (1D numpy array)
	#	coupling_array, array of coupling constants (2D numpy array)
	#	coupling_len, array of bond distance between atoms (2D numpy array)
	#	coupling_var, array of coupling constant variances (used in machine learning) (2D numpy array)

	# Define empty arrays
	shift_array = np.zeros(atoms, dtype=np.float64)
	# Variance is used for machine learning
	shift_var = np.zeros(atoms, dtype=np.float64)
	coupling_array = np.zeros((atoms, atoms), dtype=np.float64)
	coupling_len = np.zeros((atoms, atoms), dtype=np.int64)
	# Variance is used for machine learning
	coupling_var = np.zeros((atoms, atoms), dtype=np.float64)

	# Go through file looking for assignment sections
	with open(file, 'r') as f:
		shift_switch = False
		cpl_switch = False
		for line in f:
			if '> <NMREDATA_ASSIGNMENT>' in line:
				shift_switch = True
			if '> <NMREDATA_J>' in line:
				shift_switch = False
				cpl_switch = True
			# If shift assignment label found, process shift rows
			if shift_switch:
				# Shift assignment row looks like this
				#  0    , -33.56610000   , 8    , 0.00000000     \
				items = line.split()
				try:
					int(items[0])
				except:
					continue
				shift_array[int(items[0])] = float(items[2])
				shift_var[int(items[0])] = float(items[6])
			# If coupling assignment label found, process coupling rows
			if cpl_switch:
				# Coupling row looks like this
				#  0         , 4         , -0.08615310    , 3JON      , 0.00000000
				# ['0', ',', '1', ',', '-0.26456900', ',', '5JON', ',', '0.00000000']
				items = line.split()
				try:
					int(items[0])
				except:
					continue
				length = int(items[6].strip()[0])
				coupling_array[int(items[0])][int(items[2])] = float(items[4])
				coupling_var[int(items[0])][int(items[2])] = float(items[8])
				coupling_len[int(items[0])][int(items[2])] = length
				coupling_len[int(items[2])][int(items[0])] = length

	return shift_array, shift_var, coupling_array, coupling_var, coupling_len



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import copy

from autoENRICH.ml.models.model import genericmodel
from sklearn.model_selection import KFold
import sys

try:
	import tensorflow as tf
	from keras.models import Sequential
	from keras.layers import Dense, Activation, Flatten
	from keras.optimizers import RMSprop
	from keras.callbacks import EarlyStopping
except:
	print('Tensorflow not available')


class NNmodel(genericmodel):
	def __init__(self, id='NNmodel', x=[], y=[], params={}, model_args={}):
		genericmodel.__init__(self, id, x, y, params, model_args)

	def train(self):
		self.net = Sequential()

		self.net.add(Dense(int(self.params['hidden_neurons']), input_shape=self.train_x[0].shape, kernel_initializer='random_uniform'))
		self.net.add(Activation('relu'))

		if self.params['hidden_layers'] > 1:
			for layer in range(int(self.params['hidden_layers'])-1):
				self.net.add(Dense(int(self.params['hidden_neurons']), input_shape=self.train_x[0].shape, kernel_initializer='random_uniform'))
				self.net.add(Activation('relu'))
		self.net.add(Flatten())

		self.net.add(Dense(1, input_dim=int(self.params['hidden_neurons']), kernel_initializer='random_uniform'))
		optzer = RMSprop(lr=10**self.params['learning_rate'], rho=0.9)
		self.net.compile(optimizer=optzer, loss='mse')

		es = EarlyStopping(monitor='loss', mode='min', verbose=1)

		self.net.fit(self.train_x, self.train_y, epochs=int(self.params['nn_epochs']), batch_size=int(self.params['batch_size']), verbose=1, callbacks=[es])
		self.trained = True

	def predict(self, test_x):

		print(self.train_x.shape, test_x.shape)
		pred_y = self.net.predict(test_x, batch_size=None)

		return pred_y

	def cv_predict(self, fold):
		kf = KFold(n_splits=fold)
		kf.get_n_splits(self.train_x)
		pred_y = []

		for train_index, test_index in kf.split(self.train_x):

			tmp_train_x = np.asarray(self.train_x[train_index])
			tmp_train_y = self.train_y[train_index]
			tmp_test_x = np.asarray(self.train_x[test_index])
			tmp_test_y = self.train_y[test_index]

			tmp_model = copy.deepcopy(self)
			tmp_model.train_x = tmp_train_x
			tmp_model.train_y = tmp_train_y

			tmp_model.train()

			pred_y.extend(tmp_model.predict(tmp_test_x))

		pred_y = np.asarray(pred_y)

		return pred_y















###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

def check_combination(modelflag, featureflag):


	if modelflag == 'KRR':
		if featureflag not in ['CMAT', 'aSLATM', 'ACSF']:
			return False

	elif modelflag == 'FCHL':
		if featureflag != 'FCHL':
			return False


	elif modelflag == 'NN':
		return True


	elif modelflag == 'TFM':
		if featureflag != 'BCAI':
			return False

	else:
		return False


	return True



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import openbabel
from autoENRICH.reference.periodic_table import Get_periodic_table


def mol_read_type(mol):
	type_list = []
	type_array = np.zeros(len(mol.atoms), dtype=np.int32)
	Periodic_table = Get_periodic_table()
	for i in range(len(mol.atoms)):
		type = int(mol.atoms[i].atomicnum)
		type_array[i] = type
		type_list.append(Periodic_table[type])

	return type_list, type_array

def mol_getatoms(mol):
	atomnumber = len(mol.atoms)
	return atomnumber

def mol_read_xyz(mol):
	xyz_array = np.zeros((len(mol.atoms),3), dtype=np.float64)
	for i in range(len(mol.atoms)):
		xyz_array[i][0] = float(mol.atoms[i].coords[0])
		xyz_array[i][1] = float(mol.atoms[i].coords[1])
		xyz_array[i][2] = float(mol.atoms[i].coords[2])

	# Return array is zero indexed
	return xyz_array

def mol_read_dist(mol):
	xyz_array = mol_read_xyz(mol)
	atoms = mol_getatoms(mol)
	d_array = np.zeros((atoms, atoms), dtype=np.float64)
	for i in range(atoms):
		for j in range(atoms):
			d_array[i][j] = np.absolute(np.linalg.norm(xyz_array[i] - xyz_array[j]))

	return d_array



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from autoENRICH.file_creation.structure_formats import xyz
from autoENRICH.file_creation import confsearch
from autoENRICH.file_creation import HPC_submission as HPCsub
import pybel as pyb

def conformational_search(molecule, prefs, pickle_file, path=''):
	# Read relevant preferences
	iterations = prefs['conf_search']['iterations']
	maxconfs = prefs['conf_search']['maxconfs']
	RMSthresh = prefs['conf_search']['RMSthresh']
	Ethresh = prefs['conf_search']['Ethresh']

	system = prefs['comp']['system']
	python_env = prefs['comp']['python_env']
	aE_dir = prefs['comp']['aE_directory']
	memory = prefs['conf_search']['memory']
	processors = prefs['conf_search']['processors']
	walltime = prefs['conf_search']['walltime']

	# Try reading pybel mol object from xyz named according to molname
	try:
		file = path + molecule.molid + '.xyz'
		mol = next(pyb.readfile('xyz', file))
	# If not, make a molecule xyz file and then read the pybel mol object
	except:
		file = path + molecule.molid + '_INIT.xyz'
		xyz.nmrmol_to_xyz(molecule, file, num=-404)
		mol = next(pyb.readfile('xyz', file))

	smiles = mol.write("smi").split()[0]

	scriptname = 'conf_search/do_confsearch.py'
	confsearch.make_confsearch_script(scriptname, pickle_file, molecule, smiles,
									iterations=iterations, RMSthresh=RMSthresh,
									maxconfs=maxconfs, Ethresh=Ethresh, path=path)

	jobname = 'aE_' + molecule.molid + '_confsearch'
	strings = HPCsub.make_HPC_header(jobname=jobname, system=system, nodes=1, ppn=processors, walltime=walltime, mem=memory)
	strings.append('source activate {env:<10s}'.format(env=python_env))
	strings.append('python {0:<10s}'.format(scriptname))

	if prefs['comp']['system'] == 'BC3':
		filename = path + 'confsearch_' + molecule.molid + '_' + '.qsub'
	elif prefs['comp']['system'] == 'BC4':
		filename = path + 'confsearch_' + molecule.molid + '_' + '.slurm'
	elif prefs['comp']['system'] == 'localbox':
		filename = path + 'confsearch_'+ molecule.molid + '_' + '.sh'

	with open(filename, 'w') as f:
		for string in strings:
			print(string, file=f)

	print('Conformational search script + submission file produced')












###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from . import HPS_generic as generic
import numpy as np
import itertools
import pickle
import sys


def random_search(dataset, args):

	# determine whether log dictionary was provided
	if len(args['param_logs']) == 0:
		check_logs = False
	else:
		check_logs = True


	generic.setup_logfile(args)

	BEST_SCORE = 999.999
	BEST_PARAMS = {}
	next_point_to_probe = {}

	for _ in range(int(args['epochs'])):
		for param in args['param_ranges'].keys():
			if args['param_logs'][param] == 'no':
				continue
			next_point_to_probe[param] = np.random.uniform(args['param_ranges'][param][0], args['param_ranges'][param][1])

		if check_logs:
			for param in args['param_ranges'].keys():
				if args['param_logs'][param] == 'no':
					continue
				if args['param_logs'][param] == 'log':
					next_point_to_probe[param] = 10**next_point_to_probe[param]


		score, BEST_SCORE, BEST_PARAMS = generic.HPS_iteration(_, dataset, args, next_point_to_probe=next_point_to_probe,
															BEST_SCORE=BEST_SCORE, BEST_PARAMS=BEST_PARAMS)

	outname = generic.save_models(dataset, BEST_PARAMS, args)
	print('Optimised model saved in ', outname)

	return dataset, BEST_SCORE



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from autoENRICH.util.flag_handler.hdl_targetflag import target_to_flag
import numpy as np

# Attempts to determine if the molecules are the same molecule
def mol_isequal(mol1, mol2, print=False):
	# input:
	# mol1: autoENRICH nmrmol object (molecule 1)
	# mol2: autoENRICH nmrmol object (molecule 2)
	# print: True/False , whether to print information about bad matches

	# Returns: True/False

	# First simple check: same number of atoms
	if len(mol1.types) != len(mol2.types):
		if print:
			print('Molecules are different sizes')
		return False
	atoms = len(mol1.types)

	# Check for same chemical formula
	counts1 = np.zeros(1000, dtype=np.int32)
	counts2 = np.zeros(1000, dtype=np.int32)
	for i in range(atoms):
		counts1[mol1.types[i]] += 1
		counts2[mol2.types[i]] += 1
	Fail = False
	for i in range(1000):
		if counts1[i] == 0 and counts2[i] == 0:
			continue
		if counts1[i] != counts2[i]:
			Fail = True
	if Fail:
		if print:
			print('Molecules have different compositions')
		return False

	# Check for same connectivity
	# Done by checking both have the same number of each coupling type: 1JCH, 3JNH, . . .
	cpl_dict1 = {}
	cpl_dict2 = {}
	for i in range(atoms):
		for j in range(atoms):
			if mol1.conn[i][j] != 0:
				flag1 = target_to_flag([mol1.coupling_len[i][j], mol1.types[i], mol1.types[j]])
				flag2 = target_to_flag([mol2.coupling_len[i][j], mol2.types[i], mol2.types[j]])

				if flag1 not in cpl_dict1.keys():
					cpl_dict1[flag1] = 0
				if flag2 not in cpl_dict2.keys():
					cpl_dict2[flag2] = 0

				cpl_dict1[flag1] = cpl_dict1[flag1] + 1
				cpl_dict2[flag2] = cpl_dict2[flag2] + 1
	if cpl_dict1 != cpl_dict2:
		if print:
			print('Connectivity is different')
			return False

	# Return true if no checks fail
	return True


















###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

"""
autoENRICH main module
======================

Provides
	1. analysis
	2. boltzmann
	3. conformational_search
	4. file_creation
	5. file_read
	6. ml
	7. molecule
	8. preferences
	9. pybel_helpers
	10. reference
	11. top_level (command modules)
	12. util

"""

from . import analysis
from . import boltzmann
from . import conformational_search
from . import file_creation
from . import file_read
from . import ml
from . import molecule
from . import preferences
from . import pybel_helpers
from . import reference
from . import top_level
from . import util

__author__ = "Will Gerrard"
__copyright__ = "Copyright 2020"
__credits__ = ["Will Gerrard (2020) https://github.com/wg12385/autoENRICH"]
__licence__ = "GNU APGL v3+"
__version__ = "0.0.1"
__maintainer__ = "Will Gerrard"
__email__ = "will.gerrard@bristol.ac.uk"
__status__ = "Beta"



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
import pickle

from autoENRICH.conformational_search import conformational_search as conf_search
from .conformer import conformer as conformerclass
from .nmrmol import nmrmol
import glob

# global molecule object, contains everything for one auto-ENRICH run
class molecule(nmrmol):

	def __init__(self, molid, path=''):
		nmrmol.__init__(self, molid, path)

		self.conformers = []
		self.stage = 'init'

	# function to remove redundant conformers
	def remove_redundant(self, threshold=0.001):
		self.conformers = redundant_conformer_elim(self.conformers, threshold=threshold)

	# average NMR properties
	def boltzmann_average(self):
		# should there be equivalent bolzmann functions for variance ??
		self.shift = boltzmann_shift(self.conformers)
		self.coupling = boltzmann_coupling(self.conformers)
		self.variance = boltzmann_variance(self.conformers)

	# do conformational searching
	def generate_conformers(self, smiles, path='', iterations=100, RMSthresh=1, maxconfs=100, Ethresh=100000):
		xyzs, energies = conf_search.torsional_search(self, smiles, iterations=iterations, RMSthresh=RMSthresh)
		print('Initial search complete,', len(xyzs), 'conformers found')

		xyzs, energies = conf_search.select_conformers(xyzs, energies, maxconfs=maxconfs, Ethresh=Ethresh)

		for x, xyz in enumerate(xyzs):
			new_conf = conformerclass(str(x), path=path)
			new_conf.xyz = xyz
			new_conf.types = self.types
			new_conf.conn = self.conn
			new_conf.coupling_len = self.coupling_len
			new_conf.energy = energies[x]
			self.conformers.append(new_conf)

		self.stage = 'pre-opt'


###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

import sys
from rdkit import Chem
from rdkit.Chem import AllChem
import pybel as pyb
import numpy as np

# raw mass generation of xyz coordinates based on torsional angle searching
def torsional_search(smiles, iterations=100000, RMSthresh=1):
	# Input:
	# smiles: smiles string representing the molecule (string)
	# iterations: number of ETRKG iterations (integer)
	# RMSthresh: RMS threshold for ETRKG search (how different new conformers need to be) (float)

	# Returns: xyzs (list (conformers) of lists (atoms) of xyz coordinates), energies (list of conformer MMFF94s energies)

	xyzs = []
	energies = []
	# read mol into RDkit from smiles string
	rdmol = Chem.MolFromSmiles(smiles)
	# rdkit is an absolute piece of c**p, so it wont read in hydrogens, it has to add them itself
	rdmol = Chem.AddHs(rdmol)
	# Do conformational search by ETRKG
	# Riniker, S.; Landrum, G. A. “Better Informed Distance Geometry: Using What We Know To Improve Conformation Generation” J. Chem. Inf. Comp. Sci. 55:2562-74 (2015)
	ids = AllChem.EmbedMultipleConfs(rdmol,
								  clearConfs=True,
								  numConfs=iterations,
								  pruneRmsThresh=RMSthresh)
	# align conformers, not strictly neccesary but should make visualisation more convenient later on
	AllChem.AlignMolConformers(rdmol)
	# Optimise conformers by MMFF, returns success states (ignored atm) and energies
	rd_es = AllChem.MMFFOptimizeMoleculeConfs(rdmol, mmffVariant='MMFF94s')
	# Record energies in list
	for e in rd_es:
		energies.append(e[1])

	# Get list of conformer IDs
	confIds = [x.GetId() for x in rdmol.GetConformers()]
	# Define empty array for lists of coordinates
	xyzs = []
	# Loop through conformers
	for id in confIds:
		xyz = []
		# Loop over length of molecule (defined by size of mol type array)
		for t in range(len(rdmol.getatoms())):
			# append atom coordinates
			xyz.append([float(rdmol.GetConformer(id).GetAtomPosition(t)[0]),
						float(rdmol.GetConformer(id).GetAtomPosition(t)[1]),
						float(rdmol.GetConformer(id).GetAtomPosition(t)[2])])
		xyzs.append(xyz)

	return xyzs, energies


# select conformers based on coverage of the chemical space
def select_conformers(xyzs, energies, maxconfs=100, Ethresh=100000):
	# Input:
	# xyzs: list of lists of xyz coordinates from torsional search (n x m) n=number of conformers, m=number of atoms
	# energies: list of energies (n)
	# maxconfs: maximum number of conformers to return (int)
	# Ethresh: Energy threshold above which conformers will be automatically discarded (float)

	# Returns: xyzs (list (conformers) of lists (atoms) of xyz coordinates), energies (list of conformer MMFF94s energies)

	remove = []
	# Mark for removal all conformers with energy over threshold
	for i in range(len(energies)):
		if energies[i] > Ethresh:
			remove.append(i)
	# Remove marked conformers
	print('Removing', len(remove), 'conformers because MMFF energy too high')
	for id in sorted(remove, reverse=True):
		del energies[remove]
		del xyzs[remove]

	# If still too many conformers
	if len(xyzs) > maxconfs:
		# Run geomtry based conformer removal
		remove = select_over_space(xyzs, to_remove=len(xyzs)-maxconfs)
		# Remove marked conformers selected by geometric similarity
		print('Removing', len(remove), 'conformers due to similarity to other conformers')
		for id in sorted(remove, reverse=True):
			del energies[id]
			del xyzs[id]
	# Return xyz coords and energies of selected conformers
	return xyzs, energies

# xyz based selection of most similar structures
def select_over_space(xyzs, to_remove):
	#Input:
	# xyzs: list of lists of xyz coordinates from torsional search (n x m) n=number of conformers, m=number of atoms
	# to_remove: number of conformers to be removed

	# Returns: list of conformer ids to remove

	dist = np.zeros((len(xyzs),len(xyzs)), dtype=np.float64)
	remove = []
	# loop over all conformer/conformer pairs
	for aa, a in enumerate(xyzs):
		for bb, b in enumerate(xyzs):
			if aa == bb:
				continue
			# Get distance between conformers geometrically
			# Technically this is a thing called the frobeius distance/norm
			dist[aa][bb] = np.linalg.norm(np.asarray(a)-np.asarray(b))

	# Keep looping until enough conformers are marked for removal
	while len(remove) < to_remove:
		id = 0
		lowest_dist = 9999999999999999
		# Loop over conformers
		for i in range(dist.shape[0]):
			# Get sum of distances to all other conformers for conformer i
			sumdist = np.sum(dist[i])
			# Store lowest distance and conformer id
			if sumdist <= lowest_dist and i not in remove:
				id = i
				lowest_dist = sumdist
		# add least geometrically different conformer to removal list
		remove.append(id)
	# Return Ids to remove
	return remove









###



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.

from autoENRICH.reference.periodic_table import Get_periodic_table

def flag_to_target(flag):

	p_table = Get_periodic_table()
	if len(flag) == 4:
		if str(flag[0]) == 'J':
			length = int(flag[1])
		else:
			length = int(flag[0])
		atype1 = int(p_table.index(str(flag[2])))
		atype2 = int(p_table.index(str(flag[3])))

		if atype1 >= atype2:
			return [length, atype1, atype2]
		else:
			return [length, atype2, atype1]

	elif len(flag) == 3:
		atype = int(p_table.index(str(flag[0])))
		return [atype]

	else:
		print('flag, ', flag, ' not recognised, coupling flag format is <nJxy> . . .')
		print('flag, ', flag, ' not recognised, chemical shift flag format is <XCS> . . .')
		return 0

def target_to_flag(target):
	p_table = Get_periodic_table()

	if len(target) == 3:
		flag = str(target[0]) + 'J' + str(p_table[target[1]]) + str(p_table[target[2]])
	elif len(target) == 1:
		flag = str(target[0]) + 'CS'
	else:
		print('Error, target ', target, ' not recognised')

	return flag



# Copyright 2020 Will Gerrard
#This file is part of autoENRICH.

#autoENRICH is free software: you can redistribute it and/or modify
#it under the terms of the GNU Affero General Public License as published by
#the Free Software Foundation, either version 3 of the License, or
#(at your option) any later version.

#autoENRICH is distributed in the hope that it will be useful,
#but WITHOUT ANY WARRANTY; without even the implied warranty of
#MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#GNU Affero General Public License for more details.

#You should have received a copy of the GNU Affero General Public License
#along with autoENRICH.  If not, see <https://www.gnu.org/licenses/>.


import qml

import numpy as np
import copy
from qml.math import cho_solve
from qml.kernels import laplacian_kernel
from sklearn.model_selection import KFold

import tracemalloc
# top level model class

class genericmodel(object):

	def __init__(self, id='genericmodel', train_x=[], train_y=[], params={},
									model_args={}):

		self.train_x = train_x
		self.train_y = train_y

		self.params = params
		self.args = model_args

		self.trained = False


	def train(self):
		print('Stub function, why are you running this ??')

	def predict(self):
		print('Stub function, why are you running this ??')


	def cv_predict(self, fold):
		kf = KFold(n_splits=fold)
		kf.get_n_splits(self.train_x)
		pred_y = []

		for train_index, test_index in kf.split(self.train_x):

			self.train(train_x=self.train_x[train_index], train_y=self.train_y[train_index])

			pred_y.extend(self.predict(self.train_x[test_index], train_x=self.train_x[train_index]))

		pred_y = np.asarray(pred_y)

		return pred_y


























##
